import dataclasses as dc
import json
import typing as t
from abc import ABC, abstractmethod
from enum import Enum

from bson import objectid
from pymongo.change_stream import CollectionChangeStream

from superduperdb.container.serializable import Serializable

TokenType = t.Dict[str, str]


class BaseDatabaseListener(ABC):
    """
    A Base class which defines basic functions to implement.
    """

    @abstractmethod
    def listen(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def setup_cdc(self) -> CollectionChangeStream:
        pass

    @abstractmethod
    def next_cdc(self, stream: CollectionChangeStream) -> None:
        pass


class CachedTokens:
    token_path = '.cdc.tokens'
    separate = '\n'

    def __init__(self):
        # BROKEN: self._current_tokens is never read from
        self._current_tokens = []

    def append(self, token: TokenType) -> None:
        with open(CachedTokens.token_path, 'a') as fp:
            stoken = json.dumps(token)
            stoken = stoken + self.separate
            fp.write(stoken)

    def load(self) -> t.Sequence[TokenType]:
        with open(CachedTokens.token_path) as fp:
            tokens = fp.read().split(self.separate)[:-1]
            self._current_tokens = [TokenType(json.loads(t)) for t in tokens]
        return self._current_tokens


class DBEvent(str, Enum):
    """`DBEvent` simple enum to hold mongo basic events."""

    delete = 'delete'
    insert = 'insert'
    update = 'update'


class ObjectId(objectid.ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, objectid.ObjectId):
            raise TypeError('Id is required.')
        return str(v)


@dc.dataclass
class Packet:
    """
    A base packet to represent message in task queue.
    """

    ids: t.List[t.Union[ObjectId, str]]
    query: t.Optional[Serializable]

    event_type: DBEvent = DBEvent.insert

    @property
    def is_delete(self) -> bool:
        return self.event_type == DBEvent.delete

    @staticmethod
    def collate(packets: t.Sequence['Packet']) -> 'Packet':
        """
        Collate a batch of packets into one
        """

        ids = [packet.ids[0] for packet in packets]
        query = packets[0].query

        # TODO: cluster Packet for each event.
        event_type = packets[0].event_type
        return Packet(ids=ids, query=query, event_type=event_type)



from .base import config, configs, jsonable, logger
from .misc.superduper import superduper

__all__ = 'CFG', 'ICON', 'JSONable', 'ROOT', 'config', 'log', 'logging', 'superduper'

ICON = '🔮'
CFG = configs.CFG
JSONable = jsonable.JSONable
ROOT = configs.ROOT

logging = log = logger.logging

__version__ = '0.0.13'



from __future__ import annotations

import dataclasses as dc
import io
import typing as t
from contextlib import contextmanager
from functools import cached_property

import torch
from torch.utils import data
from torch.utils.data import DataLoader
from tqdm import tqdm

import superduperdb as s
from superduperdb import logging
from superduperdb.container.artifact import Artifact
from superduperdb.container.document import Document
from superduperdb.container.encoder import Encodable
from superduperdb.container.metric import Metric
from superduperdb.container.model import Model, _TrainingConfiguration
from superduperdb.container.serializable import Serializable
from superduperdb.db.base.db import DB
from superduperdb.db.base.query import Select
from superduperdb.db.query_dataset import QueryDataset
from superduperdb.ext.torch.utils import device_of, eval, to_device

if t.TYPE_CHECKING:
    from superduperdb.container.dataset import Dataset


class BasicDataset(data.Dataset):
    """
    Basic database iterating over a list of documents and applying a transformation

    :param documents: documents
    :param transform: function
    """

    def __init__(self, documents, transform):
        super().__init__()
        self.documents = documents
        self.transform = transform

    def __len__(self):
        return len(self.documents)

    def __getitem__(self, item):
        document = self.documents[item]
        if isinstance(document, Document):
            document = document.unpack()
        elif isinstance(document, Encodable):
            document = document.x
        return self.transform(document)


@dc.dataclass
class TorchTrainerConfiguration(_TrainingConfiguration):
    """
    Configuration for the PyTorch trainer.

    :param objective: Objective function
    :param loader_kwargs: Kwargs for the dataloader
    :param max_iterations: Maximum number of iterations
    :param no_improve_then_stop: Number of iterations to wait for improvement
                                 before stopping
    :param splitter: Splitter for the data
    :param download: Whether to download the data
    :param validation_interval: How often to validate
    :param listen: Which metric to listen to for early stopping
    :param optimizer_cls: Optimizer class
    :param optimizer_kwargs: Kwargs for the optimizer
    :param target_preprocessors: Preprocessors for the target

    """

    objective: t.Optional[t.Union[Artifact, t.Callable]] = None
    loader_kwargs: t.Dict = dc.field(default_factory=dict)
    max_iterations: int = 10**100
    no_improve_then_stop: int = 5
    splitter: t.Optional[Artifact] = None
    download: bool = False
    validation_interval: int = 100
    listen: str = 'objective'
    optimizer_cls: Artifact = Artifact(torch.optim.Adam, serializer='pickle')
    optimizer_kwargs: t.Dict = dc.field(default_factory=dict)
    target_preprocessors: t.Optional[t.Union[Artifact, t.Dict]] = None

    def __post_init__(self):
        if self.objective and not isinstance(self.objective, Artifact):
            self.objective = Artifact(artifact=self.objective)

        if self.target_preprocessors and not isinstance(
            self.target_preprocessors, Artifact
        ):
            self.target_preprocessors = Artifact(artifact=self.target_preprocessors)


@dc.dataclass
class Base:
    collate_fn: t.Optional[t.Union[Artifact, t.Callable]] = None
    is_batch: t.Optional[t.Union[Artifact, t.Callable]] = None
    num_directions: int = 2
    metrics: t.Optional[t.Sequence[t.Union[str, Metric]]] = None
    training_select: t.Optional[Select] = None

    @contextmanager
    def evaluating(self):
        raise NotImplementedError

    def train(self):
        raise NotImplementedError

    if t.TYPE_CHECKING:
        # TODO: We can't use the actual declarations because of the multiple
        # inheritance issues.
        @property
        def training_keys(self) -> t.List[str]:
            raise NotImplementedError

        def _validate(
            self, db: DB, validation_set: Dataset, metrics: t.Sequence[Metric]
        ):
            raise NotImplementedError

        def append_metrics(self, d: t.Dict[str, float]) -> None:
            raise NotImplementedError

    @property
    def optimizers(self) -> t.List:
        raise NotImplementedError

    def preprocess(self, r):
        raise NotImplementedError  # implemented in PyTorch wrapper and PyTorch pipeline

    def save(self, database: DB):
        raise NotImplementedError

    def stopping_criterion(self, iteration):
        max_iterations = self.training_configuration.max_iterations
        no_improve_then_stop = self.training_configuration.no_improve_then_stop
        if isinstance(max_iterations, int) and iteration >= max_iterations:
            return True
        if isinstance(no_improve_then_stop, int):
            if self.training_configuration.listen == 'objective':
                to_listen = [-x for x in self.metric_values['objective']]
            else:
                to_listen = self.metric_values[self.training_configuration.listen]

            if max(to_listen[-no_improve_then_stop:]) < max(to_listen):
                logging.info('early stopping triggered!')
                return True
        return False

    def saving_criterion(self):
        if self.training_configuration.listen == 'objective':
            to_listen = [-x for x in self.metric_values['objective']]
        else:
            to_listen = self.metric_values[self.training_configuration.listen]
        if all([to_listen[-1] >= x for x in to_listen[:-1]]):
            return True
        return False

    def _fit(
        self,
        X: t.Union[t.List[str], str],
        y: t.Optional[t.Union[t.List, t.Any]] = None,
        configuration: t.Optional[TorchTrainerConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[DB] = None,
        metrics: t.Optional[t.List[Metric]] = None,
        select: t.Optional[t.Union[Select, t.Dict]] = None,
        validation_sets: t.Optional[t.List[str]] = None,
    ):
        if configuration is not None:
            self.training_configuration = configuration
        if isinstance(select, dict):
            self.training_select = Serializable.deserialize(select)
        else:
            self.training_select = select
        if validation_sets is not None:
            self.validation_sets = validation_sets
        if metrics is not None:
            self.metrics = metrics

        self.train_X = X
        self.train_y = y

        train_data, valid_data = self._get_data(db=db)

        loader_kwargs = self.training_configuration.loader_kwargs
        train_dataloader = DataLoader(train_data, **loader_kwargs)
        valid_dataloader = DataLoader(valid_data, **loader_kwargs)

        if db is None:
            raise ValueError('db cannot be None')
        return self._fit_with_dataloaders(
            train_dataloader,
            valid_dataloader,
            db=db,
            validation_sets=validation_sets or [],
        )

    def log(self, **kwargs):
        out = ''
        for k, v in kwargs.items():
            if isinstance(v, dict):
                for kk, vv in v.items():
                    out += f'{k}/{kk}: {vv}; '
            else:
                out += f'{k}: {v}; '
        logging.info(out)

    def forward(self, X):
        return self.object.artifact(X)

    def extract_batch_key(self, batch, key: t.Union[t.List[str], str]):
        if isinstance(key, str):
            return batch[key]
        return [batch[k] for k in key]

    def extract_batch(self, batch):
        if self.train_y is not None:
            return [
                self.extract_batch_key(batch, self.train_X),
                self.extract_batch_key(batch, self.train_y),
            ]
        return [self.extract_batch_key(batch, self.train_X)]

    def take_step(self, batch, optimizers):
        batch = self.extract_batch(batch)
        outputs = self.train_forward(*batch)
        objective_value = self.training_configuration.objective.artifact(*outputs)
        for opt in optimizers:
            opt.zero_grad()
        objective_value.backward()
        for opt in optimizers:
            opt.step()
        return objective_value

    def compute_validation_objective(self, valid_dataloader):
        objective_values = []
        with self.evaluating(), torch.no_grad():
            for batch in valid_dataloader:
                batch = self.extract_batch(batch)
                objective_values.append(
                    self.training_configuration.objective.artifact(
                        *self.train_forward(*batch)
                    ).item()
                )
            return sum(objective_values) / len(objective_values)

    def _fit_with_dataloaders(
        self,
        train_dataloader: DataLoader,
        valid_dataloader: DataLoader,
        db: DB,
        validation_sets: t.List[str],
    ):
        self.train()
        iteration = 0
        while True:
            for batch in train_dataloader:
                train_objective = self.take_step(batch, self.optimizers)
                self.log(fold='TRAIN', iteration=iteration, objective=train_objective)

                if iteration % self.training_configuration.validation_interval == 0:
                    valid_loss = self.compute_validation_objective(valid_dataloader)
                    all_metrics = {}
                    assert isinstance(self.metrics, list)
                    assert all(isinstance(m, Metric) for m in self.metrics)
                    metrics = [t.cast(Metric, m) for m in self.metrics]
                    for vs in validation_sets:
                        m = self._validate(
                            db=db,
                            validation_set=vs,  # type: ignore[arg-type]
                            metrics=metrics,
                        )
                        all_metrics.update(m)
                    all_metrics.update({'objective': valid_loss})

                    self.append_metrics(all_metrics)
                    self.log(fold='VALID', iteration=iteration, **all_metrics)
                    if self.saving_criterion():
                        self.save(db)
                    stop = self.stopping_criterion(iteration)
                    if stop:
                        return
                iteration += 1

    def train_preprocess(self):
        preprocessors = {}
        if isinstance(self.train_X, str):
            preprocessors[self.train_X] = (
                self.preprocess if self.preprocess else lambda x: x
            )
        else:
            for model, X in zip(self.models, self.train_X):
                preprocessors[X] = (
                    model.preprocess if model.preprocess is not None else lambda x: x
                )
        if self.train_y is not None:
            if (
                isinstance(self.train_y, str)
                and self.training_configuration.target_preprocessors
            ):
                preprocessors[
                    self.train_y
                ] = self.training_configuration.target_preprocessors.get(
                    self.train_y, lambda x: x
                )
            elif isinstance(self.train_y, str):
                preprocessors[self.train_y] = lambda x: x
            elif (
                isinstance(self.train_y, list)
                and self.training_configuration.target_preprocessors
            ):
                for y in self.train_y:
                    preprocessors[
                        y
                    ] = self.training_configuration.target_preprocessors.get(
                        y, lambda x: x
                    )

        for k in preprocessors:
            if isinstance(preprocessors[k], Artifact):
                preprocessors[k] = preprocessors[k].artifact
        return lambda r: {k: preprocessors[k](r[k]) for k in preprocessors}

    def _get_data(self, db: t.Optional[DB]):
        if self.training_select is None:
            raise ValueError('self.training_select cannot be None')
        train_data = QueryDataset(
            select=self.training_select,
            keys=self.training_keys,
            fold='train',
            transform=self.train_preprocess(),
            db=db,
        )
        valid_data = QueryDataset(
            select=self.training_select,
            keys=self.training_keys,
            fold='valid',
            transform=self.train_preprocess(),
            db=db,
        )
        return train_data, valid_data


@dc.dataclass
class TorchModel(Base, Model):  # type: ignore[misc]
    """
    :param optimizer_state: optional optimizer state, populated automatically on reload
    :param forward_method: method to call for prediction, defaults to __call__
    :param train_forward_method: method to call for training, defaults to __call__
    """

    optimizer_state: t.Optional[Artifact] = None
    forward_method: str = '__call__'
    train_forward_method: str = '__call__'

    def __post_init__(self):
        if self.model_to_device_method:
            s.log.warn(f'{self.model_to_device_method} will be overriden with `to`')

        self.model_to_device_method = 'to'

        super().__post_init__()

        self.object.serializer = 'torch'

        if self.optimizer_state is not None:
            self.optimizer.load_state_dict(self.optimizer_state.artifact)

        self._validation_set_cache = {}

    def to(self, device):
        self.object.artifact.to(device)

    @cached_property
    def optimizer(self):
        return self.training_configuration.optimizer_cls.artifact(
            self.object.artifact.parameters(),
            **self.training_configuration.optimizer_kwargs,
        )

    @property
    def optimizers(self) -> t.List:
        return [self.optimizer]

    def save(self, database: DB):
        self.optimizer_state = Artifact(self.optimizer.state_dict(), serializer='torch')
        database.replace(object=self, upsert=True)

    @contextmanager
    def evaluating(self):
        yield eval(self)

    def train(self):
        return self.object.artifact.train()

    def eval(self):
        return self.object.artifact.eval()

    def parameters(self):
        return self.object.parameters()

    def state_dict(self):
        return self.object.state_dict()

    @contextmanager
    def saving(self):
        with super().saving():
            was_training = self.object.training
            try:
                self.object.eval()
                yield
            finally:
                if was_training:
                    self.object.train()

    def __getstate__(self):
        state = self.__dict__.copy()
        if isinstance(self.object, torch.jit.ScriptModule) or isinstance(
            self.object, torch.jit.ScriptFunction
        ):
            f = io.BytesIO()
            torch.jit.save(self.object, f)
            state['_object_bytes'] = f.getvalue()
        return state

    def __setstate__(self, state):
        keys = state.keys()
        for k in keys:
            if k != '_object_bytes':
                self.__dict__[k] = state[k]
            else:
                state.__dict__['object'] = torch.jit.load(
                    io.BytesIO(state.pop('object_bytes'))
                )

    def _predict_one(self, x):
        with torch.no_grad(), eval(self.object.artifact):
            if self.preprocess is not None:
                x = self.preprocess.artifact(x)
            x = to_device(x, device_of(self.object.artifact))
            singleton_batch = create_batch(x)
            method = getattr(self.object.artifact, self.forward_method)
            output = method(singleton_batch)
            output = to_device(output, 'cpu')
            args = unpack_batch(output)[0]
            if self.postprocess is not None:
                args = self.postprocess.artifact(args)
            return args

    def _predict(self, x, one: bool = False, **kwargs):  # type: ignore[override]
        with torch.no_grad(), eval(self.object.artifact):
            if one:
                return self._predict_one(x)
            if not isinstance(self.preprocess, Artifact):

                def func(x):
                    return x

            else:
                func = self.preprocess.artifact
            inputs = BasicDataset(x, func)
            loader = torch.utils.data.DataLoader(inputs, **kwargs)
            out = []
            for batch in tqdm(loader, total=len(loader)):
                batch = to_device(batch, device_of(self.object.artifact))

                method = getattr(self.object.artifact, self.forward_method)
                tmp = method(batch)
                tmp = to_device(tmp, 'cpu')
                tmp = unpack_batch(tmp)
                if isinstance(self.postprocess, Artifact):
                    tmp = [self.postprocess.artifact(t) for t in tmp]
                out.extend(tmp)
            return out

    def train_forward(self, X, y=None):
        X = X.to(self.device)
        if y is not None:
            y = y.to(self.device)

        method = getattr(self.object.artifact, self.train_forward_method)
        if hasattr(self.object.artifact, 'train_forward'):
            if y is None:
                return method(X)
            else:
                return method(X, y=y)
        else:
            if y is None:
                return (method(X),)
            else:
                return [method(X), y]


def unpack_batch(args):
    """
    Unpack a batch into lines of tensor output.

    :param args: a batch of model outputs

    >>> unpack_batch(torch.randn(1, 10))[0].shape
    torch.Size([10])
    >>> out = unpack_batch([torch.randn(2, 10), torch.randn(2, 3, 5)])
    >>> type(out)
    <class 'list'>
    >>> len(out)
    2
    >>> out = unpack_batch({'a': torch.randn(2, 10), 'b': torch.randn(2, 3, 5)})
    >>> [type(x) for x in out]
    [<class 'dict'>, <class 'dict'>]
    >>> out[0]['a'].shape
    torch.Size([10])
    >>> out[0]['b'].shape
    torch.Size([3, 5])
    >>> out = unpack_batch({'a': {'b': torch.randn(2, 10)}})
    >>> out[0]['a']['b'].shape
    torch.Size([10])
    >>> out[1]['a']['b'].shape
    torch.Size([10])
    """

    if isinstance(args, torch.Tensor):
        return [args[i] for i in range(args.shape[0])]

    if isinstance(args, list) or isinstance(args, tuple):
        tmp = [unpack_batch(x) for x in args]
        batch_size = len(tmp[0])
        return [[x[i] for x in tmp] for i in range(batch_size)]

    if isinstance(args, dict):
        tmp = {k: unpack_batch(v) for k, v in args.items()}
        batch_size = len(next(iter(tmp.values())))
        return [{k: v[i] for k, v in tmp.items()} for i in range(batch_size)]

    raise NotImplementedError


def create_batch(args):
    """
    Create a singleton batch in a manner similar to the PyTorch dataloader

    :param args: single data point for batching

    >>> create_batch(3.).shape
    torch.Size([1])
    >>> x, y = create_batch([torch.randn(5), torch.randn(3, 7)])
    >>> x.shape
    torch.Size([1, 5])
    >>> y.shape
    torch.Size([1, 3, 7])
    >>> d = create_batch(({'a': torch.randn(4)}))
    >>> d['a'].shape
    torch.Size([1, 4])
    """
    if isinstance(args, (tuple, list)):
        return tuple([create_batch(x) for x in args])
    if isinstance(args, dict):
        return {k: create_batch(args[k]) for k in args}
    if isinstance(args, torch.Tensor):
        return args.unsqueeze(0)
    if isinstance(args, (float, int)):
        return torch.tensor([args])
    raise TypeError('Only tensors and tuples of tensors recursively supported...')



from superduperdb.container.encoder import Encoder
from superduperdb.ext.utils import str_shape


def vector(shape):
    """
    Create an encoder for a vector (list of ints/ floats) of a given shape

    :param shape: The shape of the vector
    """
    return Encoder(
        identifier=f'vector[{str_shape(shape)}]',
        shape=shape,
        encoder=None,
        decoder=None,
    )



import hashlib
import os
import re
import signal
import sys
import typing as t
import warnings
from contextlib import contextmanager
from io import BytesIO
from multiprocessing.pool import ThreadPool

import boto3
import requests
from tqdm import tqdm

from superduperdb import CFG, logging


class TimeoutException(Exception):
    ...


def timeout_handler(signum, frame):
    raise TimeoutException()


@contextmanager
def timeout(seconds):
    old_handler = signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, old_handler)


class Fetcher:
    """
    Fetches data from a URI

    :param headers: headers to be used for download
    :param n_workers: number of download workers
    """

    def __init__(self, headers: t.Optional[t.Dict] = None, n_workers: int = 0):
        session = boto3.Session()
        self.headers = headers
        self.s3_client = session.client("s3")
        self.request_session = requests.Session()
        self.request_adapter = requests.adapters.HTTPAdapter(
            max_retries=3,
            pool_connections=n_workers if n_workers else 1,
            pool_maxsize=n_workers * 10,
        )
        self.request_session.mount("http://", self.request_adapter)
        self.request_session.mount("https://", self.request_adapter)

    def _download_s3_object(self, uri):
        f = BytesIO()
        path = uri.split('s3://')[-1]
        bucket_name = path.split('/')[0]
        file = '/'.join(path.split('/')[1:])
        self.s3_client.download_fileobj(bucket_name, file, f)
        return f.getvalue()

    def _download_file(self, path):
        path = re.split('^file://', path)[-1]
        with open(path, 'rb') as f:
            return f.read()

    def _download_from_uri(self, uri):
        return self.request_session.get(uri, headers=self.headers).content

    def __call__(self, uri: str):
        """
        Download data from a URI

        :param uri: uri to download from
        """
        if uri.startswith('file://'):
            return self._download_file(uri)
        elif uri.startswith('s3://'):
            return self._download_s3_object(uri)
        elif uri.startswith('http://') or uri.startswith('https://'):
            return self._download_from_uri(uri)
        else:
            raise NotImplementedError(f'unknown type of URI "{uri}"')


class BaseDownloader:
    """
    Base class for downloading files

    :param uris: list of uris/ file names to fetch
    :param n_workers: number of multiprocessing workers
    :param timeout: set seconds until request times out
    :param headers: dictionary of request headers passed to``requests`` package
    :param raises: raises error ``True``/``False``
    """

    def __init__(
        self,
        uris: t.List[str],
        n_workers: int = 0,
        timeout: t.Optional[int] = None,
        headers: t.Optional[t.Dict] = None,
        raises: bool = True,
    ):
        self.timeout = timeout
        self.n_workers = n_workers
        self.uris = uris
        self.headers = headers or {}
        self.raises = raises

    def go(self):
        """
        Download all files
        Uses a :py:class:`multiprocessing.pool.ThreadPool` to parallelize
                          connections.
        """
        logging.info(f'number of workers {self.n_workers}')
        prog = tqdm(total=len(self.uris))
        prog.prefix = 'downloading from uris'
        self.failed = 0
        prog.prefx = "failed: 0"

        def f(i):
            prog.update()
            try:
                if self.timeout is not None:
                    with timeout(self.timeout):
                        self._download(i)
                else:
                    self._download(i)
            except TimeoutException:
                logging.warning(f'timed out {i}')
            except Exception as e:
                if self.raises:
                    raise e
                warnings.warn(str(e))
                self.failed += 1
                prog.prefix = f"failed: {self.failed} [{e}]"

        if self.n_workers == 0:
            self._sequential_go(f)
            return

        self._parallel_go(f)

    def _check_exists_if_hybrid(self, uri):
        if uri.startswith('file://'):
            file = f'{CFG.downloads.root}/{uri.split("file://")[-1]}'
        else:
            file = f'{CFG.downloads.root}/{hashlib.sha1(uri.encode()).hexdigest()}'
        if os.path.exists(file):
            return True
        return False

    def _parallel_go(self, f):
        pool = ThreadPool(self.n_workers)
        try:
            pool.map(f, range(len(self.uris)))
        except KeyboardInterrupt:
            logging.warning("--keyboard interrupt--")
            pool.terminate()
            pool.join()
            sys.exit(1)  # Kill this subprocess so it doesn't hang

        pool.close()
        pool.join()

    def _sequential_go(self, f):
        for i in range(len(self.uris)):
            f(i)


class SaveFile:
    """
    Save file to disk.

    :param root: root directory to save files to.
    """

    def __init__(self, root: str):
        self.root = root

    def __call__(self, bytes_: bytearray, uri: str, **kwargs) -> None:
        path = f'{self.root}/{hashlib.sha1(uri.encode()).hexdigest()}'
        with open(path, 'wb') as f:
            f.write(bytes_)


class Downloader(BaseDownloader):
    """
    Download files from a list of URIs.

    :param uris: list of uris/ file names to fetch
    :param update_one: function to call to insert data into table
    :param ids: list of ids of rows/ documents to update
    :param keys: list of keys in rows/ documents to insert to
    :param n_workers: number of multiprocessing workers
    :param headers: dictionary of request headers passed to``requests`` package
    :param skip_existing: if ``True`` then don't bother getting already present data
    :param timeout: set seconds until request times out
    :param raises: raises error ``True``/``False``
    """

    results: t.Dict[int, str]

    def __init__(
        self,
        uris,
        update_one: t.Optional[t.Callable] = None,
        ids: t.Optional[t.Union[t.List[str], t.List[int]]] = None,
        keys: t.Optional[t.List[str]] = None,
        n_workers: int = 20,
        headers: t.Optional[t.Dict] = None,
        skip_existing: bool = True,
        timeout: t.Optional[int] = None,
        raises: bool = True,
    ):
        super().__init__(
            uris, n_workers=n_workers, timeout=timeout, headers=headers, raises=raises
        )

        if ids is not None:
            if len(ids) != len(uris):
                raise ValueError(f'len(ids={ids}) != len(uris={uris})')

        self.ids = ids
        self.keys = keys
        self.failed = 0
        self.skip_existing = skip_existing
        self.update_one = update_one
        self.fetcher = Fetcher(headers=headers, n_workers=n_workers)

    def _download(self, i):
        if CFG.downloads.hybrid:
            if self._check_exists_if_hybrid(self.uris[i]):
                return
        content = self.fetcher(self.uris[i])
        self.update_one(
            id=self.ids[i],
            key=self.keys[i],
            bytes_=content,
            uri=self.uris[i],
        )


def gather_uris(
    documents: t.Sequence[t.Dict], gather_ids: bool = True
) -> t.Tuple[t.List[str], t.List[str], t.Union[t.List[int], t.List[str]]]:
    """
    Get the uris out of all documents as denoted by ``{"_content": ...}``

    :param documents: list of dictionaries
    :param gather_ids: if ``True`` then gather ids of documents
    """
    uris = []
    mongo_keys = []
    ids = []
    for i, r in enumerate(documents):
        sub_uris, sub_mongo_keys = _gather_uris_for_document(r)
        if gather_ids:
            ids.extend([r['_id'] for _ in sub_uris])
        else:
            ids.append(i)
        uris.extend(sub_uris)
        mongo_keys.extend(sub_mongo_keys)
    return uris, mongo_keys, ids


def _gather_uris_for_document(r: t.Dict):
    '''
    >>> _gather_uris_for_document({'a': {'_content': {'uri': 'test'}}})
    (['test'], ['a'])
    >>> d = {'b': {'a': {'_content': {'uri': 'test'}}}}
    >>> _gather_uris_for_document(d)
    (['test'], ['b.a'])
    >>> d = {'b': {'a': {'_content': {'uri': 'test', 'bytes': b'abc'}}}}
    >>> _gather_uris_for_document(d)
    ([], [])
    '''
    uris = []
    keys = []
    for k in r:
        if isinstance(r[k], dict) and '_content' in r[k]:
            if 'uri' in r[k]['_content'] and 'bytes' not in r[k]['_content']:
                keys.append(k)
                uris.append(r[k]['_content']['uri'])
        elif isinstance(r[k], dict) and '_content' not in r[k]:
            sub_uris, sub_keys = _gather_uris_for_document(r[k])
            uris.extend(sub_uris)
            keys.extend([f'{k}.{key}' for key in sub_keys])
    return uris, keys



from __future__ import annotations

import enum
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy
import numpy.typing


class BaseVectorIndex:
    name: t.Optional[str] = None
    index: t.List[str]
    lookup: t.Dict[str, t.Union[t.Iterator[int], int]]
    measure: str

    def __init__(self, h, index, measure):
        if hasattr(h, 'tolist'):
            h = h.tolist()
        self.h_list = h
        self._h = numpy.array(h)
        self.index = index
        if index is not None:
            self.lookup = dict(zip(index, range(len(index))))
        self.measure = measure

    @property
    def h(self):
        if len(self.h_list) != self._h.shape[0]:
            self._h = numpy.array(self.h_list)
        return self._h

    @property
    def shape(self):
        return self.h.shape

    def find_nearest_from_id(self, _id, n=100):
        _ids, scores = self.find_nearest_from_ids([_id], n=n)
        return _ids[0], scores[0]

    def find_nearest_from_ids(self, _ids, n=100):
        ix = list(map(self.lookup.__getitem__, _ids))
        return self.find_nearest_from_arrays(self.h[ix, :], n=n)

    def find_nearest_from_array(self, h, n=100):
        h = to_numpy(h)
        _ids, scores = self.find_nearest_from_arrays(h[None, :], n=n)
        return _ids[0], scores[0]

    def find_nearest_from_arrays(self, h, n=100):
        raise NotImplementedError

    def __getitem__(self, item):
        raise NotImplementedError


def to_numpy(x: numpy.typing.ArrayLike) -> numpy.ndarray:
    return numpy.array(x)


class VectorIndexMeasureType(str, enum.Enum):
    cosine = 'cosine'
    css = 'css'
    dot = 'dot'
    l2 = 'l2'


VectorCollectionId = str
VectorCollectionItemId = str
VectorIndexMeasure = t.Union[VectorIndexMeasureType, str]


class VectorCollectionItemNotFound(Exception):
    pass


@dataclass(frozen=True)
class VectorCollectionConfig:
    id: VectorCollectionId
    dimensions: int
    measure: VectorIndexMeasure = VectorIndexMeasureType.l2
    parameters: t.Mapping[str, t.Any] = field(default_factory=dict)


@dataclass(frozen=True)
class VectorCollectionItem:
    id: VectorCollectionItemId
    vector: numpy.ndarray

    @classmethod
    def create(
        cls,
        *,
        id: VectorCollectionItemId,
        vector: numpy.typing.ArrayLike,
    ) -> VectorCollectionItem:
        return VectorCollectionItem(id=id, vector=to_numpy(vector))

    def to_dict(self) -> t.Dict:
        return {'id': self.id, 'vector': self.vector}


@dataclass(frozen=True)
class VectorCollectionResult:
    id: VectorCollectionItemId
    score: float


class VectorCollection(ABC):
    """A vector collection within a vector database.

    Concrete implementations of this class are responsible for lifecycle management of
    a single vector collection within a specific vector database.

    It is assumed that a vector collection is associated with a single vector column or
    field within a vector database.
    """

    @abstractmethod
    def add(self, items: t.Sequence[VectorCollectionItem]) -> None:
        """
        Add items to the collection.

        :param items: items to add
        """
        pass

    @abstractmethod
    def find_nearest_from_id(
        self,
        identifier: VectorCollectionItemId,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the item with the given identifier.

        :param identifier: identifier of the item
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """
        pass

    @abstractmethod
    def find_nearest_from_array(
        self,
        array: numpy.typing.ArrayLike,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the given vector.

        :param array: array representing the vector
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """
        pass


class VectorDatabase(ABC):
    """
    A vector database that manages vector collections.

    Concrete implementations of this class are responsible for lifecycle management of
    vector collections within a specific vector database.

    Implementations of this abstract base class retrieve vector collections by their
    identifier for subsequent operations.
    """

    @abstractmethod
    def get_table(self, config: VectorCollectionConfig) -> VectorCollection:
        """Get a vector collection by its identifier.

        :param config: configuration for the vector database
        """
        pass



import dataclasses as dc
import typing as t

from superduperdb.data.tree.for_each import for_each

from .typed_cache import TypedCache

ContentType = t.TypeVar('ContentType')


@dc.dataclass
class Cached(t.Generic[ContentType]):
    """A base class for classes that are cached using their uri and type"""

    _content: dc.InitVar[ContentType] = None
    uri: str = ''

    @property
    def content(self) -> ContentType:
        """Get the cached content. `content` is not JSONized"""
        return self._content  # type: ignore[attr-defined]

    @content.setter
    def content(self, content: ContentType):
        """Set the cached content. `content` is not JSONized"""
        self._content = content  # type: ignore[attr-defined]

    def __post_init__(self, content: ContentType):
        self.content = content


class URICache(TypedCache):
    """
    A typed cache for instances of `Cached`.
    """

    def cache(self, entry: t.Any) -> None:
        """Add a uri to the content in an instance of Cached"""
        if isinstance(entry, Cached):
            entry.uri = self.put(entry.content)

    def uncache(self, entry: t.Any) -> None:
        """Retrieve the content from the uri in an instance of Cached"""
        if isinstance(entry, Cached):
            entry.content = self.get(entry.uri)

    def cache_all(self, entry: t.Any) -> None:
        """Run self.cache recursively"""
        for_each(self.cache, entry)

    def uncache_all(self, entry: t.Any) -> None:
        """Run self.uncache recursively"""
        for_each(self.uncache, entry)



import dataclasses as dc
import datetime
import json
import threading
import traceback
import typing as t
from collections import Counter
from enum import Enum

from pymongo.change_stream import CollectionChangeStream

from superduperdb import logging
from superduperdb.db.base.db import DB
from superduperdb.db.mongodb import CDC_COLLECTION_LOCKS, query
from superduperdb.misc.runnable.runnable import Event

from .base import BaseDatabaseListener, CachedTokens, DBEvent, Packet, TokenType
from .handler import CDC_QUEUE, CDCHandler

MongoChangePipelines: t.Dict[str, t.Sequence[t.Any]] = {'generic': []}


class CDCKeys(str, Enum):
    """
    A enum to represent mongo change document keys.
    """

    operation_type = 'operationType'
    document_key = 'documentKey'
    update_descriptions_key = 'updateDescription'
    update_field_key = 'updatedFields'
    document_data_key = 'fullDocument'
    deleted_document_data_key = 'documentKey'


class _DatabaseListenerThreadScheduler(threading.Thread):
    def __init__(
        self,
        listener: BaseDatabaseListener,
        stop_event: Event,
        start_event: Event,
    ) -> None:
        threading.Thread.__init__(self, daemon=True)
        self.stop_event = stop_event
        self.start_event = start_event
        self.listener = listener

    def run(self) -> None:
        try:
            cdc_stream = self.listener.setup_cdc()
            self.start_event.set()
            while not self.stop_event.is_set():
                self.listener.next_cdc(cdc_stream)
        except Exception:
            logging.error('In _DatabaseListenerThreadScheduler')
            traceback.print_exc()


@dc.dataclass
class MongoChangePipeline:
    """`MongoChangePipeline` is a class to represent listen pipeline
    in mongodb watch api.
    """

    matching_operations: t.Sequence[str] = dc.field(default_factory=list)

    def validate(self):
        raise NotImplementedError

    def build_matching(self) -> t.Sequence[t.Dict]:
        """A helper fxn to build a listen pipeline for mongo watch api.

        :param matching_operations: A list of operations to watch.
        """
        if bad := [op for op in self.matching_operations if op not in DBEvent]:
            raise ValueError(f'Unknown operations: {bad}')

        return [{'$match': {'operationType': {'$in': [*self.matching_operations]}}}]


class MongoEventMixin:
    """A Mixin class which defines helper fxns for `MongoDatabaseListener`
    It define basic events handling methods like
    `on_create`, `on_update`, etc.
    """

    DEFAULT_ID: str = '_id'
    EXCLUSION_KEYS: t.Sequence[str] = [DEFAULT_ID]

    def on_create(self, change: t.Dict, db: DB, collection: query.Collection) -> None:
        """on_create.
        A helper on create event handler which handles inserted document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param change: The changed document.
        :param db: a superduperdb instance.
        :param collection: The collection on which change was observed.
        """
        logging.debug('Triggered `on_create` handler.')
        # new document added!
        document = change[CDCKeys.document_data_key]
        ids = [document[self.DEFAULT_ID]]
        cdc_query = collection.find()
        packet = Packet(ids=ids, event_type=DBEvent.insert, query=cdc_query)
        CDC_QUEUE.put_nowait(packet)

    def on_update(self, change: t.Dict, db: DB, collection: query.Collection) -> None:
        """on_update.

        A helper on update event handler which handles updated document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param change: The changed document.
        :param db: a superduperdb instance.
        :param collection: The collection on which change was observed.
        """

        # TODO: Handle removed fields and updated fields.
        document = change[CDCKeys.document_key]
        ids = [document[self.DEFAULT_ID]]
        cdc_query = collection.find()
        packet = Packet(ids=ids, event_type=DBEvent.insert, query=cdc_query)
        CDC_QUEUE.put_nowait(packet)

    def on_delete(self, change: t.Dict, db: DB, collection: query.Collection) -> None:
        """on_delete.

        A helper on delete event handler which handles deleted document in the
        change stream.
        It basically extracts the change document and build the taskflow graph to
        execute.

        :param change: The changed document.
        :param db: a superduperdb instance.
        :param collection: The collection on which change was observed.
        """
        logging.debug('Triggered `on_delete` handler.')
        # new document added!
        document = change[CDCKeys.deleted_document_data_key]
        ids = [document[self.DEFAULT_ID]]
        packet = Packet(ids=ids, event_type=DBEvent.delete, query=None)
        CDC_QUEUE.put_nowait(packet)


class MongoDatabaseListener(BaseDatabaseListener, MongoEventMixin):
    """
    It is a class which helps capture data from mongodb database and handle it
    accordingly.

    This class accepts options and db instance from user and starts a scheduler
    which could schedule a listening service to listen change stream.

    This class builds a workflow graph on each change observed.

    """

    IDENTITY_SEP: str = '/'
    _scheduler: t.Optional[threading.Thread]

    _change_pipeline: t.Union[str, t.Sequence[t.Dict], None] = None

    def __init__(
        self,
        db: DB,
        on: query.Collection,
        stop_event: Event,
        identifier: 'str' = '',
        resume_token: t.Optional[TokenType] = None,
    ):
        """__init__.

        :param db: It is a superduperdb instance.
        :param on: It is used to define a Collection on which CDC would be performed.
        :param stop_event: A threading event flag to notify for stoppage.
        :param identifier: A identifier to represent the listener service.
        :param resume_token: A resume token is a token used to resume
        the change stream in mongo.
        """
        self.db = db
        self._on_component = on
        self._identifier = self._build_identifier([identifier, on.identifier])
        self.tokens = CachedTokens()
        self._change_counters = Counter(inserts=0, updates=0, deletes=0)
        self.resume_token = None

        if resume_token is not None:
            self.resume_token = resume_token

        self._change_pipeline = None
        self._stop_event = stop_event
        self._startup_event = Event()
        self._scheduler = None
        self.start_handler()

        super().__init__()

    @property
    def identity(self) -> str:
        return self._identifier

    def start_handler(self):
        self._cdc_change_handler = CDCHandler(db=self.db, stop_event=self._stop_event)
        self._cdc_change_handler.start()

    @classmethod
    def _build_identifier(cls, identifiers) -> str:
        """_build_identifier.

        :param identifiers: list of identifiers.
        :rtype: str
        """
        return cls.IDENTITY_SEP.join(identifiers)

    @staticmethod
    def _get_stream_pipeline(
        change: str,
    ) -> t.Optional[t.Sequence[t.Any]]:
        """_get_stream_pipeline.

        :param change: change can be a prebuilt listen pipeline like
        'generic' or user Defined listen pipeline.
        """
        return MongoChangePipelines.get(change)

    def _get_reference_id(self, document: t.Dict) -> t.Optional[str]:
        """_get_reference_id.

        :param document:
        """
        try:
            document_key = document[CDCKeys.document_key]
            reference_id = str(document_key['_id'])
        except KeyError:
            return None
        return reference_id

    def event_handler(self, change: t.Dict) -> None:
        """event_handler.
        A helper fxn to handle incoming changes from change stream on a collection.

        :param change: The change (document) observed during the change stream.
        """
        event = change[CDCKeys.operation_type]
        reference_id = self._get_reference_id(change)

        if not reference_id:
            logging.warn('Document change not handled due to no document key')
            return

        if event == DBEvent.insert:
            self._change_counters['inserts'] += 1
            self.on_create(change, db=self.db, collection=self._on_component)

        elif event == DBEvent.update:
            self._change_counters['updates'] += 1
            self.on_update(change, db=self.db, collection=self._on_component)

        elif event == DBEvent.delete:
            self._change_counters['deletes'] += 1
            self.on_delete(change, db=self.db, collection=self._on_component)

    def dump_token(self, change: t.Dict) -> None:
        """dump_token.
        A helper utility to dump resume token from the changed document.

        :param change:
        """
        token = change[self.DEFAULT_ID]
        self.tokens.append(token)

    def check_if_taskgraph_change(self, change: t.Dict) -> bool:
        """
        A helper method to check if the cdc change is done
        by taskgraph nodes.
        """
        if change[CDCKeys.operation_type] == DBEvent.update:
            updates = change[CDCKeys.update_descriptions_key]
            updated_fields = updates[CDCKeys.update_field_key]
            return any(
                [True for k in updated_fields.keys() if k.startswith('_outputs')]
            )
        return False

    def setup_cdc(self) -> CollectionChangeStream:
        """
        Setup cdc change stream from user provided
        """
        if isinstance(self._change_pipeline, str):
            pipeline = self._get_stream_pipeline(self._change_pipeline)

        elif isinstance(self._change_pipeline, list):
            pipeline = self._change_pipeline or None

        else:
            raise TypeError(
                'Change pipeline can be either a string or a dictionary, '
                f'provided {type(self._change_pipeline)}'
            )
        stream = self._on_component.change_stream(
            pipeline=pipeline, resume_after=self.resume_token
        )

        stream_iterator = stream(self.db)
        logging.info(f'Started listening database with identity {self.identity}...')
        return stream_iterator

    def next_cdc(self, stream: CollectionChangeStream) -> None:
        """
        Get the next stream of change observed on the given `Collection`.
        """

        change = stream.try_next()
        if change is not None:
            logging.debug(f'Database change encountered at {datetime.datetime.now()}')

            if not self.check_if_taskgraph_change(change):
                self.event_handler(change)
            self.dump_token(change)

    def attach_scheduler(self, scheduler: threading.Thread) -> None:
        self._scheduler = scheduler

    def info(self) -> t.Dict:
        """
        Get info on the current state of listener.
        """
        info = {}
        info.update(
            {
                'inserts': self._change_counters['inserts'],
                'updates': self._change_counters['updates'],
                'deletes': self._change_counters['deletes'],
            }
        )
        logging.info(json.dumps(info, indent=2))
        return info

    def set_resume_token(self, token: TokenType) -> None:
        """
        Set the resume token for the listener.
        """
        self.resume_token = token

    def set_change_pipeline(
        self, change_pipeline: t.Optional[t.Union[str, t.Sequence[t.Dict]]]
    ) -> None:
        """
        Set the change pipeline for the listener.
        """
        if change_pipeline is None:
            change_pipeline = MongoChangePipelines.get('generic')

        self._change_pipeline = change_pipeline

    def resume(self, token: TokenType) -> None:
        """
        Resume the listener from a given token.
        """
        self.set_resume_token(token)
        self.listen()

    def listen(
        self,
        change_pipeline: t.Optional[t.Union[str, t.Sequence[t.Dict]]] = None,
        timeout: t.Optional[float] = None,
    ) -> None:
        """Primary fxn to initiate listening of a database on the collection
        with defined `change_pipeline` by the user.

        :param change_pipeline: A mongo listen pipeline defined by the user
        for more fine grained listening.
        """
        try:
            CDC_COLLECTION_LOCKS[self._on_component.identifier] = True

            self._stop_event.clear()
            if self._scheduler:
                if self._scheduler.is_alive():
                    raise RuntimeError(
                        'CDC Listener thread is already running!,/'
                        'Please stop the listener first.'
                    )

            if not self._cdc_change_handler.is_alive():
                del self._cdc_change_handler
                self.start_handler()

            scheduler = _DatabaseListenerThreadScheduler(
                self, stop_event=self._stop_event, start_event=self._startup_event
            )
            self.attach_scheduler(scheduler)
            self.set_change_pipeline(change_pipeline)
            assert self._scheduler is not None
            self._scheduler.start()

            self._startup_event.wait(timeout=timeout)
        except Exception:
            logging.error('Listening service stopped!')
            CDC_COLLECTION_LOCKS.pop(self._on_component.identifier, None)

            self.stop()
            raise

    def last_resume_token(self) -> TokenType:
        """
        Get the last resume token from the change stream.
        """
        return self.tokens.load()[0]

    def resume_tokens(self) -> t.Sequence[TokenType]:
        """
        Get the resume tokens from the change stream.
        """
        return self.tokens.load()

    def stop(self) -> None:
        """
        Stop listening cdc changes.
        This stops the corresponding services as well.
        """
        CDC_COLLECTION_LOCKS.pop(self._on_component.identifier, None)

        self._stop_event.set()
        self._cdc_change_handler.join()
        if self._scheduler:
            self._scheduler.join()

    def is_available(self) -> bool:
        """
        Get the status of listener.
        """
        return not self._stop_event.is_set()

    def close(self) -> None:
        self.stop()



import os
import typing as t

import lance
import numpy as np
import pyarrow as pa

from superduperdb.vector_search.base import (
    VectorCollection,
    VectorCollectionConfig,
    VectorCollectionItem,
    VectorCollectionItemId,
    VectorCollectionResult,
    VectorDatabase,
)

# TODO: Implement a vector index manager for ``lance`` after tidying
# base classes and interfaces in ``superduperdb.vector_search.base``.


class LanceVectorDatabase(VectorDatabase):
    """A class for managing multiple Lance datasets."""

    def __init__(self, root_dir: str) -> None:
        """
        :param root_dir: root directory that contains all Lance datasets
        """
        assert root_dir.startswith('lance://')
        self.root_dir = root_dir.split('lance://')[1]

    # TODO: ``create`` kwarg is not included in the ABC interface but it is assumed part
    # of the interface in ``superduperdb/container/vector_index.py``! Short-term fix
    # is to include it here but long-term fix is to tidy up the ABC interface.
    def get_table(
        self, config: VectorCollectionConfig, create: bool = False
    ) -> VectorCollection:
        """Retrieve a managed Lance dataset according to the given config."""
        uri = os.path.join(self.root_dir, config.id)
        return LanceVectorCollection(
            uri=uri, dimensions=config.dimensions, measure=config.measure
        )


class LanceVectorCollection(VectorCollection):
    """A class for managing a lance dataset."""

    def __init__(self, uri: str, dimensions: int, measure: str = 'cosine') -> None:
        """
        :param uri: URI of the Lance dataset
        :param dimensions: dimension of the vector embeddings in the Lance dataset
        :param measure: measure which defines strategy to use for vector search.
        """
        self.uri = uri
        self.dimensions = dimensions
        self.measure = measure

    @property
    def _dataset(self) -> lance.LanceDataset:
        try:
            return lance.dataset(self.uri)
        except Exception as e:
            raise Exception(f'Could not load Lance dataset at {self.uri}') from e

    # TODO: ``upsert`` kwarg is not included in the ABC interface but it is assumed part
    # of the interface in ``superduperdb/db/mongodb/cdc/vector_task_factory.py``!
    # Short-term fix is to include it here, long-term fix is to tidy up ABC interface.

    def size(self) -> int:
        """
        Get the number of rows in lance dataset.
        """
        return self._dataset.count_rows()

    def delete_from_ids(self, ids: t.Sequence[str]) -> None:
        """
        Delete vectors from the lance dataset.
        :param ids: t.Sequence of identifiers.
        """
        to_remove = ", ".join(f"'{str(id)}'" for id in ids)
        self._dataset.delete(f"id IN ({to_remove})")

    def add(self, data: t.Sequence[VectorCollectionItem], upsert: bool = False) -> None:
        """
        Add vectors to existing Lance dataset or create a new one if it doesn't exist.

        :param data: t.Sequence of ``VectorCollectionItem`` objects.
        """

        # TODO: Handle floating point bits
        typ = pa.list_(
            pa.field('values', pa.float32(), nullable=False), self.dimensions
        )

        vectors = []
        for d in data:
            vector = d.vector
            if hasattr(vector, 'numpy'):
                vector = vector.numpy()
            vectors.append(vector)

        _vecs = pa.array([vector for vector in vectors], type=typ)
        _ids = pa.array([item.id for item in data], type=pa.string())
        _table = pa.Table.from_arrays([_ids, _vecs], names=['id', 'vector'])

        if not os.path.exists(self.uri):
            if upsert:
                lance.write_dataset(_table, self.uri, mode='create')
            else:
                raise FileNotFoundError('URI is either invalid or does not exists')
        else:
            lance.write_dataset(_table, self.uri, mode='append')

        # TODO: Add logic to optimise Lance dataset for vector search at
        # regular intervals or after a certain number of items have been added.
        # These optimisations include fragment merging, compacting, deleting, etc.
        # See ``lance.DatasetOptimizer`` for further details.

    def find_nearest_from_id(
        self,
        identifier: VectorCollectionItemId,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the item with the given identifier.

        The ``lance`` file format has been specifically designed for fast
        random access. The logic to take advantage of this is implemented
        by the ``.take`` method.

        See https://blog.lancedb.com/benchmarking-random-access-in-lance-ed690757a826
        for further details.

        :param identifier: identifier of the item
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """

        # ``.take`` expect a list of integers or pyarrow array as input
        vector = self._dataset.take([int(identifier)], columns=['vector']).to_pydict()[
            'vector'
        ][0]
        return self.find_nearest_from_array(
            vector, limit=limit, offset=offset, within_ids=within_ids
        )

    def find_nearest_from_array(
        self,
        array: np.typing.ArrayLike,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
        measure: t.Optional[str] = None,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the given vector.

        :param array: array representing the vector
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """

        # NOTE: filter is currently applied AFTER vector-search
        # See https://lancedb.github.io/lance/api/python/lance.html#lance.dataset.LanceDataset.scanner
        if within_ids:
            assert (
                type(within_ids) == tuple
            ), 'within_ids must be a tuple for lance sql parser'
            result = self._dataset.to_table(
                columns=['id'],
                nearest={
                    'column': 'vector',
                    'q': array,
                    'k': limit,
                    'metric': measure if measure else self.measure,
                },
                filter=f"id in {within_ids}",
                offset=offset,
            )
        else:
            result = self._dataset.to_table(
                columns=['id'],
                nearest={"column": 'vector', "q": array, "k": limit},
                offset=offset,
            )

        ids = result['id'].to_pylist()
        scores = result['_distance'].to_pylist()

        return [
            VectorCollectionResult(id=id_, score=(-1 * distance))
            for id_, distance in zip(ids, scores)
        ]



import dataclasses as dc
import re
import typing as t

from superduperdb.db.base.cursor import SuperDuperCursor


@dc.dataclass
class SuperDuperIbisCursor(SuperDuperCursor):
    db: t.Any = None

    def _get_encoders_for_schema(self, schema_names):
        encoders = {}
        for name in schema_names:
            try:
                # TODO use version
                identifier, version = re.match(
                    '.*::_encodable=(.*)/(.*)::', name
                ).groups()[:2]
            except AttributeError as e:
                if "'NoneType' object has no attribute 'groups'" in str(e):
                    continue
                raise e
            name = name.split('::_encodable')[0]
            encoders[name] = self.encoders[identifier]
        return encoders

    def execute(self):
        if self.raw_cursor[0] is None:
            return

        schema_names = self.raw_cursor.schema().names

        self.raw_cursor = self.raw_cursor.execute()
        self.encoders = self._get_encoders_for_schema(schema_names)

        if self.encoders:
            new_cols = [c.split('::_encodable')[0] for c in self.raw_cursor.columns]
            self.raw_cursor.columns = new_cols

        self.dict_cursor = self.raw_cursor.to_dict(orient="records")
        self._n = len(self.dict_cursor)
        self._index = 0
        return self

    def cursor_next(self):
        if self._index < self._n:
            row = self.dict_cursor[self._index]
            self._index += 1
            return row
        else:
            raise StopIteration



import typing as t

from superduperdb import CFG, logging
from superduperdb.container.document import Document
from superduperdb.container.serializable import Serializable
from superduperdb.db.base.download import Downloader, SaveFile, gather_uris
from superduperdb.db.base.query import Insert, Select


def download_content(
    db,
    query: t.Union[Select, Insert, t.Dict],
    ids: t.Optional[t.Sequence[str]] = None,
    documents: t.Optional[t.List[Document]] = None,
    timeout: t.Optional[int] = None,
    raises: bool = True,
    n_download_workers: t.Optional[int] = None,
    headers: t.Optional[t.Dict] = None,
    download_update: t.Optional[t.Callable] = None,
    **kwargs,
) -> t.Optional[t.Sequence[Document]]:
    """
    Download content contained in uploaded data. Items to be downloaded are identifier
    via the subdocuments in the form exemplified below. By default items are downloaded
    to the database, unless a ``download_update`` function is provided.

    :param db: database instance
    :param query: query to be executed
    :param ids: ids to be downloaded
    :param documents: documents to be downloaded
    :param timeout: timeout for download
    :param raises: whether to raise errors
    :param n_download_workers: number of download workers
    :param headers: headers to be used for download
    :param download_update: function to be used for updating the database
    :param **kwargs: additional keyword arguments

    >>> d = {"_content": {"uri": "<uri>", "encoder": "<encoder-identifier>"}}
    >>> def update(key, id, bytes):
    >>> ... with open(f'/tmp/{key}+{id}', 'wb') as f:
    >>> ...     f.write(bytes)
    >>> download_content(None, None, ids=["0"], documents=[d]))
    ...
    """
    logging.debug(query)
    logging.debug(ids)
    update_db = False
    if isinstance(query, dict):
        query = Serializable.deserialize(query)

    if documents is not None:
        pass
    elif isinstance(query, Select):
        update_db = True
        if ids is None:
            documents = list(db.execute(query).raw_cursor)
        else:
            select = query.select_using_ids(ids)
            documents = list(db.execute(select))
    else:
        assert isinstance(query, Insert)
        documents = t.cast(t.List[Document], query.documents)

    uris, keys, place_ids = gather_uris([d.encode() for d in documents])
    if uris:
        logging.info(f'found {len(uris)} uris')
    if not uris:
        return None

    if n_download_workers is None:
        try:
            n_download_workers = db.metadata.get_metadata(key='n_download_workers')
        except TypeError:
            n_download_workers = 0

    if headers is None:
        try:
            headers = db.metadata.get_metadata(key='headers')
        except TypeError:
            pass

    if timeout is None:
        try:
            timeout = db.metadata.get_metadata(key='download_timeout')
        except TypeError:
            pass

    if CFG.downloads.hybrid:
        _download_update = SaveFile(CFG.downloads.root)
    else:

        def _download_update(key, id, bytes_, **kwargs):  # type: ignore[misc]
            return query.download_update(db=db, key=key, id=id, bytes=bytes_)

    assert place_ids is not None
    downloader = Downloader(
        uris=uris,
        ids=place_ids,
        keys=keys,
        update_one=download_update or _download_update,
        n_workers=n_download_workers,
        timeout=timeout,
        headers=headers,
        raises=raises,
    )
    downloader.go()
    if update_db:
        return None

    for id_, key in zip(place_ids, keys):
        documents[id_] = db.db.set_content_bytes(  # type: ignore[call-overload]
            documents[id_],  # type: ignore[call-overload]
            key,
            downloader.results[id_],  # type: ignore[index]
        )
    return documents



import dataclasses as dc
import itertools
import typing as t

import tqdm
from overrides import override

import superduperdb as s
from superduperdb import logging
from superduperdb.container.component import Component
from superduperdb.container.document import Document
from superduperdb.container.encoder import Encodable
from superduperdb.container.listener import Listener
from superduperdb.db.base.db import DB
from superduperdb.db.mongodb import CDC_COLLECTION_LOCKS
from superduperdb.misc.special_dicts import MongoStyleDict
from superduperdb.vector_search.base import (
    VectorCollectionConfig,
    VectorCollectionItem,
    VectorIndexMeasureType,
)

if t.TYPE_CHECKING:
    pass

T = t.TypeVar('T')


def ibatch(iterable: t.Iterable[T], batch_size: int) -> t.Iterator[t.List[T]]:
    """
    Batch an iterable into chunks of size `batch_size`

    :param iterable: the iterable to batch
    :param batch_size: the number of groups to write
    """
    iterator = iter(iterable)
    while True:
        batch = list(itertools.islice(iterator, batch_size))
        if not batch:
            break
        yield batch


@dc.dataclass
class VectorIndex(Component):
    """
    A component carrying the information to apply a vector index to a ``DB`` instance

    :param identifier: Unique string identifier of index
    :param indexing_listener: Listener which is applied to created vectors
    :param compatible_listener: Listener which is applied to vectors to be compared
    :param measure: Measure to use for comparison
    :param version: version of this index
    :param metric_values: Metric values for this index
    """

    identifier: str
    indexing_listener: t.Union[Listener, str]
    compatible_listener: t.Union[None, Listener, str] = None
    measure: VectorIndexMeasureType = VectorIndexMeasureType.cosine
    version: t.Optional[int] = None
    metric_values: t.Optional[t.Dict] = dc.field(default_factory=dict)

    type_id: t.ClassVar[str] = 'vector_index'

    @override
    def on_create(self, db: DB) -> None:
        if s.CFG.vector_search == s.CFG.data_backend:
            if (create := getattr(db.databackend, 'create_vector_index', None)) is None:
                msg = 'VectorIndex is not supported by the current database backend'
                raise ValueError(msg)

            create(self)

    @override
    def on_load(self, db: DB) -> None:
        if isinstance(self.indexing_listener, str):
            self.indexing_listener = t.cast(
                Listener, db.load('listener', self.indexing_listener)
            )

        if isinstance(self.compatible_listener, str):
            self.compatible_listener = t.cast(
                Listener, db.load('listener', self.compatible_listener)
            )

        if not s.CFG.vector_search == s.CFG.data_backend:
            assert db.vector_database
            self.vector_table = db.vector_database.get_table(
                VectorCollectionConfig(
                    id=self.identifier,
                    dimensions=self.dimensions,
                    measure=self.measure,
                ),
                create=True,  # type: ignore[call-arg]
            )

            clt = self.indexing_listener.select.table_or_collection

            assert isinstance(clt.identifier, str), 'clt.identifier must be a string'

            if not CDC_COLLECTION_LOCKS.get(clt.identifier, False):
                self._initialize_vector_database(db)

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        out = [('indexing_listener', 'listener')]
        if self.compatible_listener is not None:
            out.append(('compatible_listener', 'listener'))
        return out

    def get_vector(
        self,
        like: Document,
        models: t.List[str],
        keys: t.List[str],
        db: t.Any = None,
        outputs: t.Optional[t.Dict] = None,
        featurize: bool = True,
    ):
        document = MongoStyleDict(like.unpack())
        if featurize:
            outputs = outputs or {}
            if '_outputs' not in document:
                document['_outputs'] = {}
            document['_outputs'].update(outputs)

            assert not isinstance(self.indexing_listener, str)
            features = self.indexing_listener.features or ()
            for subkey in features:
                subout = document['_outputs'].setdefault(subkey, {})
                f_subkey = features[subkey]
                if f_subkey not in subout:
                    subout[f_subkey] = db.models[f_subkey]._predict(document[subkey])
                document[subkey] = subout[f_subkey]
        available_keys = list(document.keys()) + ['_base']
        try:
            model_name, key = next(
                (m, k) for m, k in zip(models, keys) if k in available_keys
            )
        except StopIteration:
            raise Exception(
                f'Keys in provided {like} don\'t match'
                f'VectorIndex keys: {keys}, with model: {models}'
            )

        model_input = document
        if key == '_base' and key in document:
            model_input = document[key]
        elif key != '_base':
            model_input = document[key]

        model = db.models[model_name]
        return model.predict(model_input, one=True), model.identifier, key

    def get_nearest(
        self,
        like: Document,
        db: t.Any = None,
        id_field: str = '_id',
        outputs: t.Optional[t.Dict] = None,
        featurize: bool = True,
        ids: t.Optional[t.Sequence[str]] = None,
        n: int = 100,
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """Given a document, find the nearest results in this vector index, returned as
        two parallel lists of result IDs and scores

        :param like: The document to compare against
        :param db: The datastore to use
        :param outputs: An optional dictionary
        :param featurize: Enable featurization
        :param ids: A list of ids to match
        :param n: Number of items to return
        """

        models, keys = self.models_keys
        if len(models) != len(keys):
            raise ValueError(f'len(model={models}) != len(keys={keys})')
        within_ids = ids or ()

        if isinstance(like.content, dict) and id_field in like.content:
            nearest = self.vector_table.find_nearest_from_id(
                str(like[id_field]), within_ids=within_ids, limit=n
            )
            return (
                [result.id for result in nearest],
                [result.score for result in nearest],
            )
        h = self.get_vector(
            like=like,
            models=models,
            keys=keys,
            db=db,
            outputs=outputs,
            featurize=featurize,
        )[0]
        nearest = self.vector_table.find_nearest_from_array(
            h, within_ids=within_ids, limit=n
        )
        return (
            [result.id for result in nearest],
            [result.score for result in nearest],
        )

    @property
    def models_keys(self) -> t.Tuple[t.List[str], t.List[str]]:
        """
        Return a list of model and keys for each listener
        """
        assert not isinstance(self.indexing_listener, str)
        assert not isinstance(self.compatible_listener, str)

        if self.compatible_listener:
            listeners = [self.indexing_listener, self.compatible_listener]
        else:
            listeners = [self.indexing_listener]

        models = [w.model.identifier for w in listeners]  # type: ignore[union-attr]
        keys = [w.key for w in listeners]
        return models, keys

    def _initialize_vector_database(self, db: DB) -> None:
        logging.info(f'loading hashes: {self.identifier!r}')
        assert not isinstance(self.indexing_listener, str)

        if self.indexing_listener.select is None:
            raise ValueError('.select must be set')

        progress = tqdm.tqdm(desc='Loading vectors into vector-table...')
        for record_batch in ibatch(
            db.execute(self.indexing_listener.select),
            s.CFG.cluster.backfill_batch_size,
        ):
            items = []
            for record in record_batch:
                key = self.indexing_listener.key
                if key.startswith('_outputs.'):
                    key = key.split('.')[1]

                id = record[db.databackend.id_field]
                assert not isinstance(self.indexing_listener.model, str)
                h = record.outputs(key, self.indexing_listener.model.identifier)
                if isinstance(h, Encodable):
                    h = h.x
                items.append(VectorCollectionItem.create(id=str(id), vector=h))

            self.vector_table.add(items)
            progress.update(len(items))

    @property
    def dimensions(self) -> int:
        assert not isinstance(self.indexing_listener, str)
        assert not isinstance(self.indexing_listener.model, str)
        if shape := getattr(self.indexing_listener.model.encoder, 'shape', None):
            return shape[-1]
        raise ValueError('Couldn\'t get shape of model outputs from model encoder')



import dataclasses as dc
import typing as t
from threading import Lock

from .key_cache import KeyCache

SEP = '-'


@dc.dataclass
class TypedCache:
    """Cache objects by class.

    Each class of object is given a unique name and its own cache.

    The key for an object is that unique class name, joined with the object's key from
    its class cache.
    """

    def put(self, entry: t.Any, key: t.Optional[str] = None) -> str:
        """Put an item into the cache, return a string key"""
        name = self._get_name(type(entry))
        cache = self._name_to_cache[name]
        if key is None:
            key = cache.put(entry)
            return f'{name}{SEP}{key}'

        k, _, rest = key.partition(SEP)
        if k == name and rest:
            cache.put(entry, rest)
            return key

        raise ValueError(f'Bad key {key}, expected {name}-')

    def _get_name(self, cls: t.Any) -> str:
        with self._lock:
            try:
                return self._class_to_name[cls]
            except KeyError:
                name = cls.__name__
                if name in self._name_to_cache:
                    name = f'{cls.__module__}.{name}'
                cache = KeyCache[cls]()

                self._class_to_name[cls] = name
                self._name_to_cache[name] = cache
        return name

    def get(self, key: str) -> t.Any:
        """Given a key, returns an entry or raises KeyError"""
        name, key = key.split(SEP)
        return self._name_to_cache[name].get(key)

    def expire(self, before: float) -> t.Dict[t.Type, t.Dict[str, t.Any]]:
        cn = self._class_to_name.items()
        cc = ((cls, self._name_to_cache[name]) for cls, name in cn)
        return {cls: cache.expire(before) for cls, cache in cc}

    def __contains__(self, key: str) -> bool:
        k, _, rest = key.partition(SEP)
        return rest in self._name_to_cache.get(k, {})

    def __len__(self) -> int:
        with self._lock:
            return sum(len(i) for i in self._name_to_cache.values())

    _class_to_name: t.Dict[t.Type, str] = dc.field(default_factory=dict)
    _lock: Lock = dc.field(default_factory=Lock)
    _name_to_cache: t.Dict[str, KeyCache] = dc.field(default_factory=dict)



import collections
import traceback
import uuid

from flask import Flask, jsonify, make_response, request

from superduperdb import CFG
from superduperdb.container.document import load_bson, load_bsons
from superduperdb.container.serializable import Serializable
from superduperdb.db.base.artifact import ArtifactStore
from superduperdb.db.base.build import build_datalayer
from superduperdb.misc.serialization import serializers


def make_endpoints(app, db):
    cache = collections.defaultdict(lambda: {})

    @app.route('/select', methods=['GET'])
    def select():
        """
        Execute a Select query in the database for multiple components and return
        the ID for the container of the components in the server cache.

        This endpoint deserializes the query, executes the query, generates a unique
        ID for the result, stores the result in the server cache, and returns the ID
        to the client.

        :param request_id: unique identifier for the request, generated by client
        :param query: serialized query
        """
        d = request.get_json()
        query = Serializable.deserialize(d['query'])
        output = list(db.execute(query))  # TODO - support cursor with flask streaming
        file_id = str(uuid.uuid4())
        cache[d['request_id']][file_id] = output.dump_bson()
        return jsonify({'file_id': file_id})

    @app.route('/insert', methods=['POST'])
    def insert():
        """
        Execute an Insert query in the database.

        This endpoint assumes that serialized documents have already been sent
        by the client and exist in memory in the server cache. It then deserializes
        the documents in the cache, deletes the documents from the cache and
        deserializes the query. Finally, it executes the query.

        :param request_id: unique identifier for the request, generated by client
        :param documents: unique identifier for the documents, generated by client
        :param query: serialized query
        """
        d = request.get_json()
        binary = cache[d['request_id']][d['documents']]
        del cache[d['request_id']]
        documents = load_bsons(binary, encoders=db.encoders)
        query = Serializable.deserialize(d['query'])
        query.documents = documents
        db.execute(query)
        return jsonify({'msg': 'ok'})

    @app.route('/update', methods=['POST'])
    def update():
        """
        Execute an Update query in the database.

        This endpoint assumes that serialized documents have already been sent
        by the client and exist in memory in the server cache. It then deserializes
        the documents in the cache, deletes the documents from the cache and
        deserializes the query. Finally, it executes the query.

        :param request_id: unique identifier for the request, generated by client
        :param documents: unique identifier for the documents, generated by client
        :param query: serialized query
        """
        d = request.get_json()
        binary = cache[d['request_id']][d['update']]
        del cache[d['request_id']]
        update = load_bson(binary, encoders=db.encoders)
        query = Serializable.deserialize(d['query'])
        query.update = update
        db.execute(query)
        return jsonify({'msg': 'ok'})

    @app.route('/delete', methods=['POST'])
    def delete():
        """
        Execute a delete query in the database.

        :param query: serialized delete query
        """
        d = request.get_json()['query']
        query = Serializable.deserialize(d)
        db.execute(query)
        return jsonify({'msg': 'ok'})

    @app.route('/add', methods=['POST'])
    def add():
        """
        Add a component to the database.

        This endpoint assumes that serialized artifacts have already been sent
        by the client and exist in memory in the server cache. It then deserializes
        the artifacts in the cache, adds these artifacts back to the original component,
        deserializes any metadata that is contained on the component, and then adds the
        component to the database. Finally, it deletes the artifacts from the cache.

        :param request_id: unique identifier for the request, generated by client
        :param serializers: mapping from random string to serializer name (eg `pickle`)
        :param component: serialized component
        """
        d = request.get_json()
        artifact_cache = cache[d['request_id']]
        for a in artifact_cache:
            artifact_cache[a] = serializers[d['serializers'][a]].decode(
                artifact_cache[a]
            )
        d = ArtifactStore.load_from_cache(d, cache=artifact_cache)

        object = Serializable.deserialize(d['component'])
        db.add(object)
        del cache[d['request_id']]
        return jsonify({'msg': 'ok'})

    @app.route('/artifacts/get/<request_id>/<file_id>', methods=['GET'])
    def get(request_id, file_id):
        """
        Return data originally produced by the client to the client.

        It also deletes the data from the cache.

        :param request_id: unique identifier for the request, generated by client
        :param file_id: unique identifier for the component, generated by client
        """
        bytes = cache[request_id][file_id]
        del cache[request_id][file_id]
        if not cache[request_id]:
            del cache[request_id]
        return make_response(bytes)

    @app.route('/artifacts/put/<request_id>/<file_id>', methods=['PUT'])
    def put(request_id, file_id):
        """
        Read the buffered incoming data from the client into one bytes object and
        then hold in memory in a cache.

        Artifacts are serialized on the client and then sent to the server in
        individual requests.

        :param request_id: unique identifier for the request, generated by client
        :param file_id: unique identifier for the component, generated by client
        """
        cache[request_id][file_id] = request.get_data(cache=False)
        return jsonify({'msg': 'ok'})

    @app.route('/load', methods=['GET'])
    def load():
        """
        Load component artifacts from the database into server cache and send
        component metadata to the client.

        This endpoint loads the component from the database using the provided
        query parameters, serializes the component, stores the component artifacts
        in the server cache, and sends the component metadata to the client.

        :param type_id: type_id of component to load ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param request_id: unique identifier for the request, generated by client
        :param version: (optional) numerical version - specify for full metadata
        """
        d = request.get_json()
        request_id = d['request_id']
        version = d.get('version')
        if version:
            version = int(version)
        m = db.load(type_id=d['type_id'], identifier=d['identifier'], version=version)
        serialized, artifacts = m.serialized
        lookup = {a.sha1: str(uuid.uuid4()) for a in artifacts}
        s_lookup = {lookup[a.sha1]: a.serializer for a in artifacts}
        to_send = ArtifactStore.replace_artifacts_with_dict(serialized, lookup)
        for a in artifacts:
            cache[request_id][lookup[a.sha1]] = serializers[
                s_lookup[lookup[a.sha1]]
            ].encode(a.artifact)
        return jsonify(to_send)

    @app.route('/remove', methods=['POST'])
    def remove():
        """
        Force-remove a component from the database.

        :param version: (optional) numerical version of the component to delete
        :param type_id: type_id of component to show `['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']`
        :param identifier: identifying string to component
        """
        d = request.get_json()
        version = d.get('version')
        if version:
            version = int(version)

        db.remove(
            type_id=d['type_id'],
            identifier=d['identifier'],
            version=version,
            force=True,
        )
        return jsonify({'msg': 'ok'})

    @app.route('/show', methods=['GET'])
    def show():
        """
        Show components already added to the database.

        :param type_id: type_id of component to show ['encoder', 'model', 'listener',
                       'learning_task', 'training_configuration', 'metric',
                       'vector_index', 'job']
        :param identifier: identifying string to component
        :param version: (optional) numerical version - specify for full metadata
        """
        return jsonify(db.show(**request.get_json()))

    @app.errorhandler(500)
    def handle_500(error):
        error_traceback = traceback.format_exc()
        response = {
            'traceback': error_traceback,
            'error': str(error),
            'type': str(type(error)),
        }
        return jsonify(response), 500


def make_flask_app(db):
    app = Flask(__name__)
    make_endpoints(app, db)
    return app


def serve(db=None):
    if db is None:
        db = build_datalayer()
    app = make_flask_app(db)
    app.run(CFG.server.host, CFG.server.port)


if __name__ == '__main__':
    serve()



import numpy

from superduperdb import logging
from superduperdb.vector_search.base import BaseVectorIndex


class VanillaVectorIndex(BaseVectorIndex):
    """
    Simple hash-set for looking up with vector similarity.

    :param h: ``torch.Tensor``
    :param index: list of IDs
    :param measure: measure to assess similarity
    """

    name = 'vanilla'

    def __init__(self, h, index, measure='cosine'):
        if isinstance(measure, str):
            measure = measures[measure]
        super().__init__(h, index, measure)

    def find_nearest_from_arrays(self, h, n=100):
        similarities = self.measure(h, self.h)  # mypy: ignore
        logging.debug(similarities)
        scores = -numpy.sort(-similarities, axis=1)[:, :n]
        ix = numpy.argsort(-similarities, axis=1)[:, :n]
        ix = ix.tolist()
        scores = scores.tolist()
        _ids = [[self.index[i] for i in sub] for sub in ix]
        return _ids, scores

    def __getitem__(self, item):
        ix = [self.lookup[i] for i in item]
        return VanillaVectorIndex(self.h[ix], item, self.measure)


def l2(x, y):
    return numpy.array([-numpy.linalg.norm(x - y, axis=1)])


def dot(x, y):
    return numpy.dot(x, y.T)


def cosine(x, y):
    x = x / numpy.linalg.norm(x, axis=1)[:, None]
    y = y / numpy.linalg.norm(y, axis=1)[:, None]
    return dot(x, y)


measures = {'cosine': cosine, 'dot': dot, 'l2': l2}



import typing as t
from contextlib import contextmanager

import numpy
from readerwriterlock import rwlock

from .base import (
    VectorCollection,
    VectorCollectionConfig,
    VectorCollectionId,
    VectorCollectionItem,
    VectorCollectionItemId,
    VectorCollectionItemNotFound,
    VectorCollectionResult,
    VectorDatabase,
    VectorIndexMeasure,
    VectorIndexMeasureType,
    to_numpy,
)
from .table_scan import VanillaVectorIndex


class InMemoryVectorCollection(VectorCollection):
    """
    An in-memory vector collection.

    :param dimensions: The number of dimensions of the vectors in the collection.
    :param measure: The distance measure to use for vector search.
    """

    def __init__(
        self,
        *,
        dimensions: int,
        measure: VectorIndexMeasure = VectorIndexMeasureType.l2,
    ) -> None:
        super().__init__()
        self._index = VanillaVectorIndex(
            numpy.empty((0, dimensions), dtype='float32'),
            [],
            measure,
        )
        self._lock = rwlock.RWLockFair()

    @contextmanager
    def init(self) -> t.Iterator["VectorCollection"]:
        yield self

    def add(self, items: t.Sequence[VectorCollectionItem], **kwargs) -> None:
        """
        Add items to the collection.
        """
        for item in items:
            with self._lock.gen_wlock():
                self._add(item)

    def _add(self, item: VectorCollectionItem) -> None:
        ix = self._index.lookup.get(item.id)
        if ix is not None:
            self._index.h[ix] = item.vector
        else:
            self._index.index.append(item.id)
            self._index.lookup[item.id] = len(self._index.lookup)
            self._index.h_list.append(list(item.vector))

    def find_nearest_from_id(
        self,
        identifier: VectorCollectionItemId,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the item with the given identifier.

        :param identifier: identifier of the item
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """
        with self._lock.gen_rlock():
            try:
                index = self._index if not within_ids else self._index[within_ids]
                ids, scores = index.find_nearest_from_id(identifier, n=limit + offset)
            except KeyError as e:
                raise VectorCollectionItemNotFound() from e
            return self._convert_ids_scores_to_results(ids[offset:], scores[offset:])

    def find_nearest_from_array(
        self,
        array: numpy.typing.ArrayLike,
        *,
        within_ids: t.Sequence[VectorCollectionItemId] = (),
        limit: int = 100,
        offset: int = 0,
    ) -> t.List[VectorCollectionResult]:
        """
        Find items that are nearest to the given vector.

        :param array: array representing the vector
        :param within_ids: identifiers to search within
        :param limit: maximum number of nearest items to return
        :param offset: offset of the first item to return
        """

        with self._lock.gen_rlock():
            index = self._index if not within_ids else self._index[within_ids]
            ids, scores = index.find_nearest_from_array(
                to_numpy(array),
                n=limit + offset,
            )
            return self._convert_ids_scores_to_results(ids[offset:], scores[offset:])

    def _convert_ids_scores_to_results(
        self, ids: numpy.ndarray, scores: numpy.ndarray
    ) -> t.List[VectorCollectionResult]:
        results: t.List[VectorCollectionResult] = []
        for id, score in zip(ids, scores):
            results.append(VectorCollectionResult(id=id, score=score))
        return results


class InMemoryVectorDatabase(VectorDatabase):
    """
    An in-memory vector database.
    """

    def __init__(self, uri: str = 'inmemory://') -> None:
        assert uri.startswith('inmemory://')
        self._collections: t.Dict[VectorCollectionId, VectorCollection] = {}

    def get_table(self, config: VectorCollectionConfig, **kwargs) -> VectorCollection:
        collection = self._collections.get(config.id)
        if not collection:
            collection = self._collections[config.id] = InMemoryVectorCollection(
                dimensions=config.dimensions, measure=config.measure
            )
        return collection



import dataclasses as dc
import typing as t

from superduperdb.container.artifact import Artifact
from superduperdb.container.component import Component


@dc.dataclass
class Metric(Component):
    """
    Metric base object with which to evaluate performance on a data-set.
    These objects are ``callable`` and are applied row-wise to the data, and averaged.

    :param identifier: unique identifier
    :param object: callable or ``Artifact`` to be applied to the data
    :param version: version of the ``Metric``
    """

    identifier: str
    object: t.Union[Artifact, t.Callable, None] = None
    version: t.Optional[int] = None

    artifacts: t.ClassVar[t.List[str]] = ['object']
    type_id: t.ClassVar[str] = 'metric'

    def __post_init__(self) -> None:
        if self.object and not isinstance(self.object, Artifact):
            self.object = Artifact(artifact=self.object)

    def __call__(self, x: t.Sequence[int], y: t.Sequence[int]) -> bool:
        assert isinstance(self.object, Artifact)
        return self.object.artifact(x, y)



import queue
import threading
import traceback

from superduperdb import logging
from superduperdb.container.job import FunctionJob
from superduperdb.container.task_workflow import TaskWorkflow
from superduperdb.db.base.db import DB
from superduperdb.misc.runnable.queue_chunker import QueueChunker
from superduperdb.misc.runnable.runnable import Event

from .base import Packet
from .vector_task_factory import vector_task_factory

queue_chunker = QueueChunker(chunk_size=100, timeout=0.2)
CDC_QUEUE: queue.Queue = queue.Queue()


class CDCHandler(threading.Thread):
    """
    This class is responsible for handling the change by executing the taskflow.
    This class also extends the task graph by adding funcation job node which
    does post model executiong jobs, i.e `copy_vectors`.
    """

    def __init__(self, db: DB, stop_event: Event):
        """__init__.

        :param db: a superduperdb instance.
        :param stop_event: A threading event flag to notify for stoppage.
        """
        self.db = db
        self._stop_event = stop_event

        threading.Thread.__init__(self, daemon=False)

    def run(self):
        try:
            for c in queue_chunker(CDC_QUEUE, self._stop_event):
                _submit_task_workflow(self.db, Packet.collate(c))

        except Exception as exc:
            traceback.print_exc()
            logging.info(f'Error while handling cdc batches :: reason {exc}')


def _submit_task_workflow(db: DB, packet: Packet) -> None:
    """
    Build a taskflow and execute it with changed ids.

    This also extends the task workflow graph with a node.
    This node is responsible for applying a vector indexing listener,
    and copying the vectors into a vector search database.
    """
    if packet.is_delete:
        workflow = TaskWorkflow(db)
    else:
        workflow = db._build_task_workflow(packet.query, ids=packet.ids, verbose=False)

    task = 'delete' if packet.is_delete else 'copy'
    task_callable, task_name = vector_task_factory(task=task)

    serialized_query = packet.query.serialize() if packet.query else None

    def add_node(identifier):
        from superduperdb.container.vector_index import VectorIndex

        vi = db.load(identifier=identifier, type_id='vector_index')
        assert isinstance(vi, VectorIndex)

        assert not isinstance(vi.indexing_listener, str)
        listener_id = vi.indexing_listener.identifier

        args = [listener_id, serialized_query, packet.ids]
        job = FunctionJob(callable=task_callable, args=args)
        workflow.add_node(f'{task_name}({listener_id})', job=job)

        return listener_id

    listener_ids = [add_node(i) for i in db.show('vector_index')]
    if not packet.is_delete:
        assert listener_ids
        listener_id = listener_ids[-1]
        model, _, key = listener_id.rpartition('/')
        workflow.add_edge(f'{model}.predict({key})', f'{task_name}({listener_id})')

    workflow.run_jobs()



import os
import typing as t
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

from . import config_dicts
from .config import Config

File = t.Union[Path, str]

# The top-level directory of the project
ROOT = Path(__file__).parents[2]

# The default prefix used for config environment variables
PREFIX = 'SUPERDUPERDB_'

# The name of the environment variable used to read the config files.
# This value needs to be read before all the other config values are.
FILES_NAME = 'CONFIG_FILES'

# The base name of the configs file
CONFIG_FILE = 'configs.json'

_LOCAL_CONFIG = Path(CONFIG_FILE)
_PROJECT_CONFIG = ROOT / CONFIG_FILE
_USER_CONFIG = Path(f'~/.superduperdb/{CONFIG_FILE}').expanduser()

_ALL_CONFIGS = _PROJECT_CONFIG, _LOCAL_CONFIG, _USER_CONFIG

FILE_SEP = ','


@dataclass(frozen=True)
class ConfigSettings:
    """
    A class that reads a Pydantic class from config files and environment variables.

    :param cls: The Pydantic class to read.
    :param default_files: The default config files to read.
    :param prefix: The prefix to use for environment variables.
    :param environ: The environment variables to read from.
    """

    cls: t.Type
    default_files: t.Union[t.Sequence[Path], str]
    prefix: str
    environ: t.Optional[t.Dict] = None

    @cached_property
    def config(self) -> t.Any:
        """Read a Pydantic class"""
        env = dict(os.environ if self.environ is None else self.environ)

        files = env.pop(self.prefix + FILES_NAME, self.default_files)
        if isinstance(files, str):
            files = files.split(FILE_SEP)

        parent = self.cls().dict()
        kwargs = config_dicts.config_dicts(files, parent, self.prefix, env)
        return self.cls(**kwargs)


def build_config() -> Config:
    """
    Build the config object from the environment variables and config files.
    """
    CONFIG = ConfigSettings(Config, _ALL_CONFIGS, PREFIX)
    return CONFIG.config


CFG = build_config()



import typing as t

from superduperdb.container.serializable import Serializable
from superduperdb.vector_search.base import VectorCollectionConfig, VectorCollectionItem


def delete_vectors(
    indexing_listener_identifier: str,
    cdc_query: t.Optional[Serializable],
    ids: t.Sequence[str],
    db=None,
):
    """
    A helper fxn to copy vectors of a `indexing_listener` component/model of
    a `vector_index` listener.

    This function will be added as node to the taskworkflow after every
    `indexing_listener` in the defined listeners in db.

    :param indexing_listener_identifier: A identifier of indexing listener.
    :param cdc_query: A query which will be used by `db._build_task_workflow` method
    :param ids: List of ids which were observed as changed documents.
    :param db: A ``DB`` instance.
    """
    config = VectorCollectionConfig(id=indexing_listener_identifier, dimensions=0)
    table = db.vector_database.get_table(config)
    table.delete_from_ids(ids)


def copy_vectors(
    indexing_listener_identifier: str,
    cdc_query: Serializable,
    ids: t.Sequence[str],
    db=None,
):
    """
    A helper fxn to copy vectors of a `indexing_listener` component/model of
    a `vector_index` listener.

    This function will be added as node to the taskworkflow after every
    `indexing_listener` in the defined listeners in db.

    :param indexing_listener_identifier: A identifier of indexing listener.
    :param cdc_query: A query which will be used by `db._build_task_workflow` method
    :param ids: List of ids which were observed as changed documents.
    :param db: A ``DB`` instance.
    """
    query = Serializable.deserialize(cdc_query)
    select = query.select_using_ids(ids)
    docs = db.select(select)
    docs = [doc.unpack() for doc in docs]
    model, _, key = indexing_listener_identifier.rpartition('/')
    vectors = [
        {'vector': doc['_outputs'][key][model], 'id': str(doc['_id'])} for doc in docs
    ]
    dimensions = len(vectors[0]['vector'])
    config = VectorCollectionConfig(
        id=indexing_listener_identifier, dimensions=dimensions
    )
    table = db.vector_database.get_table(config, create=True)

    vector_list = [VectorCollectionItem(**vector) for vector in vectors]
    table.add(vector_list, upsert=True)


def vector_task_factory(task: str = 'copy') -> t.Tuple[t.Callable, str]:
    if task == 'copy':
        return copy_vectors, 'copy_vectors'
    elif task == 'delete':
        return delete_vectors, 'delete_vectors'
    raise NotImplementedError(f'Unknown task: {task}')



import abc
import dataclasses as dc
import time
import typing as t
from threading import Lock

Entry = t.TypeVar('Entry')


class Cache(abc.ABC):
    def put(self, entry: t.Any) -> str:
        raise NotImplementedError

    def get(self, key: str) -> t.Any:
        raise NotImplementedError


@dc.dataclass
class KeyCache(t.Generic[Entry], Cache):
    def put(self, entry: Entry, key: t.Optional[str] = None) -> str:
        """Put an item into the cache, return a string key"""
        with self._lock:
            try:
                return self._inverse[entry]
            except KeyError:
                pass

            if key is None:
                key = str(self._count)
                self._count += 1
            elif key in self:
                raise ValueError(f'Tried to insert duplicate key {key}')

            self._cache[key] = entry, time.time()
            self._inverse[entry] = key

            return key

    def get(self, key: str) -> Entry:
        """Given a key, returns an entry or raises KeyError"""
        with self._lock:
            return self._cache[key][0]

    def expire(self, before: float) -> t.Dict[str, t.Any]:
        with self._lock:
            old = {k: e for k, (e, time) in self._cache.items() if time < before}
            for key, entry in old.items():
                self._cache.pop(key)
                self._inverse.pop(entry)
            return old

    def __contains__(self, key: str) -> bool:
        return key in self._cache

    def __len__(self) -> int:
        return len(self._cache)

    _cache: t.Dict[str, t.Tuple[Entry, float]] = dc.field(default_factory=dict)
    _count: int = 0
    _inverse: t.Dict[Entry, str] = dc.field(default_factory=dict)
    _lock: Lock = dc.field(default_factory=Lock)



import asyncio
import base64
import dataclasses as dc
import itertools
import os
import typing as t

import aiohttp
import requests
import tqdm
from openai import Audio, ChatCompletion, Embedding, Image, Model as OpenAIModel
from openai.error import RateLimitError, ServiceUnavailableError, Timeout, TryAgain

import superduperdb as s
from superduperdb.container.component import Component
from superduperdb.container.encoder import Encoder
from superduperdb.container.model import PredictMixin
from superduperdb.ext.vector.encoder import vector
from superduperdb.misc.compat import cache
from superduperdb.misc.retry import Retry

retry = Retry(
    exception_types=(RateLimitError, ServiceUnavailableError, Timeout, TryAgain)
)


def init_fn():
    s.log.info('Setting OpenAI api-key...')
    os.environ['OPENAI_API_KEY'] = s.CFG.apis.providers['openai'].api_key


@cache
def _available_models():
    return tuple([r['id'] for r in OpenAIModel.list()['data']])


@dc.dataclass
class OpenAI(Component, PredictMixin):
    """OpenAI predictor.

    :param model: The model to use, e.g. ``'text-embedding-ada-002'``.
    :param identifier: The identifier to use, e.g. ``'my-model'``.
    :param version: The version to use, e.g. ``0`` (leave empty)
    :param takes_context: Whether the model takes context into account.
    :param encoder: The encoder identifier.
    """

    model: str
    identifier: str = ''
    version: t.Optional[int] = None
    takes_context: bool = False
    encoder: t.Union[Encoder, str, None] = None
    model_update_kwargs: dict = dc.field(default_factory=dict)

    #: A unique name for the class
    type_id: t.ClassVar[str] = 'model'

    @property
    def child_components(self):
        if self.encoder is not None:
            return [('encoder', 'encoder')]
        return []

    def __post_init__(self):
        # dall-e is not currently included in list returned by OpenAI model endpoint
        if self.model not in (mo := _available_models()) and self.model not in (
            'dall-e'
        ):
            msg = f'model {self.model} not in OpenAI available models, {mo}'
            raise ValueError(msg)

        self.identifier = self.identifier or self.model

        if 'OPENAI_API_KEY' not in os.environ:
            raise ValueError('OPENAI_API_KEY not set')


@dc.dataclass
class OpenAIEmbedding(OpenAI):
    """OpenAI embedding predictor

    :param shape: The shape as ``tuple`` of the embedding.
    """

    shape: t.Optional[t.Sequence[int]] = None

    shapes: t.ClassVar[t.Dict] = {'text-embedding-ada-002': (1536,)}

    def __post_init__(self):
        super().__post_init__()
        if self.shape is None:
            self.shape = self.shapes[self.identifier]
        self.encoder = vector(self.shape)

    @retry
    def _predict_one(self, X: str, **kwargs):
        e = Embedding.create(input=X, model=self.identifier, **kwargs)
        return e['data'][0]['embedding']

    @retry
    async def _apredict_one(self, X: str, **kwargs):
        e = await Embedding.acreate(input=X, model=self.identifier, **kwargs)
        return e['data'][0]['embedding']

    @retry
    def _predict_a_batch(self, texts: t.List[str], **kwargs):
        out = Embedding.create(input=texts, model=self.identifier, **kwargs)
        return [r['embedding'] for r in out['data']]

    @retry
    async def _apredict_a_batch(self, texts: t.List[str], **kwargs):
        out = await Embedding.acreate(input=texts, model=self.identifier, **kwargs)
        return [r['embedding'] for r in out['data']]

    def _predict(self, X, one: bool = False, **kwargs):
        if isinstance(X, str):
            return self._predict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(self, X, one: bool = False, **kwargs):
        if isinstance(X, str):
            return await self._apredict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        # Note: we submit the async requests in serial to avoid rate-limiting
        for i in range(0, len(X), batch_size):
            out.extend(await self._apredict_a_batch(X[i : i + batch_size], **kwargs))
        return out


@dc.dataclass
class OpenAIChatCompletion(OpenAI):
    """OpenAI chat completion predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    takes_context: bool = True
    prompt: str = ''

    def _format_prompt(self, context, X):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt + X

    @retry
    def _predict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = self._format_prompt(context, X)
        return ChatCompletion.create(
            messages=[{'role': 'user', 'content': X}],
            model=self.identifier,
            **kwargs,
        )['choices'][0]['message']['content']

    @retry
    async def _apredict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = self._format_prompt(context, X)
        return (
            await ChatCompletion.acreate(
                messages=[{'role': 'user', 'content': X}],
                model=self.identifier,
                **kwargs,
            )
        )['choices'][0]['message']['content']

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        return [self._predict_one(msg) for msg in X]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        return [await self._apredict_one(msg) for msg in X]


@dc.dataclass
class OpenAIImageCreation(OpenAI):
    """OpenAI image creation predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    takes_context: bool = True
    prompt: str = ''

    def _format_prompt(self, context, X):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt + X

    @retry
    def _predict_one(
        self,
        X,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        if context is not None:
            X = self._format_prompt(context, X)
        if response_format == 'b64_json':
            b64_json = Image.create(prompt=X, n=n, response_format='b64_json')['data'][
                0
            ]['b64_json']
            return base64.b64decode(b64_json)
        else:
            url = Image.create(prompt=X, n=n, **kwargs)['data'][0]['url']
            return requests.get(url).content

    @retry
    async def _apredict_one(
        self,
        X,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        if context is not None:
            X = self._format_prompt(context, X)
        if response_format == 'b64_json':
            b64_json = (await Image.acreate(prompt=X, n=n, response_format='b64_json'))[
                'data'
            ][0]['b64_json']
            return base64.b64decode(b64_json)
        else:
            url = (await Image.acreate(prompt=X, n=n, **kwargs))['data'][0]['url']
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    return await resp.read()

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(
                X, n=1, response_format=response_format, context=context, **kwargs
            )
        return [
            self._predict_one(msg, n=1, response_format=response_format) for msg in X
        ]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(
                X, context=context, n=1, response_format=response_format, **kwargs
            )
        return [
            await self._apredict_one(msg, n=1, response_format=response_format)
            for msg in X
        ]


@dc.dataclass
class OpenAIImageEdit(OpenAI):
    """OpenAI image edit predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    takes_context: bool = True
    prompt: str = ''

    def _format_prompt(self, context):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt

    @retry
    def _predict_one(
        self,
        image: t.BinaryIO,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        mask_png_path: t.Optional[str] = None,
        **kwargs,
    ):
        if context is not None:
            self.prompt = self._format_prompt(context)

        if mask_png_path is not None:
            with open(mask_png_path, 'rb') as f:
                mask = f.read()
        else:
            mask = None

        if response_format == 'b64_json':
            b64_json = Image.create_edit(
                image=image,
                mask=mask,
                prompt=self.prompt,
                n=n,
                response_format='b64_json',
            )['data'][0]['b64_json']
            return base64.b64decode(b64_json)
        else:
            url = Image.create_edit(
                image=image, mask=mask, prompt=self.prompt, n=n, **kwargs
            )['data'][0]['url']
            return requests.get(url).content

    @retry
    async def _apredict_one(
        self,
        image: t.BinaryIO,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        mask_png_path: t.Optional[str] = None,
        **kwargs,
    ):
        if context is not None:
            self.prompt = self._format_prompt(context)

        if mask_png_path is not None:
            with open(mask_png_path, 'rb') as f:
                mask = f.read()
        else:
            mask = None

        if response_format == 'b64_json':
            b64_json = (
                await Image.acreate_edit(
                    image=image,
                    mask=mask,
                    prompt=self.prompt,
                    n=n,
                    response_format='b64_json',
                )
            )['data'][0]['b64_json']
            return base64.b64decode(b64_json)
        else:
            url = (
                await Image.acreate_edit(
                    image=image, mask=mask, prompt=self.prompt, n=n, **kwargs
                )
            )['data'][0]['url']
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    return await resp.read()

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(
                image=X, n=1, response_format=response_format, context=context, **kwargs
            )
        return [
            self._predict_one(
                image=image, n=1, response_format=response_format, **kwargs
            )
            for image in X
        ]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(
                image=X, context=context, n=1, response_format=response_format, **kwargs
            )
        return [
            await self._apredict_one(
                image=image, n=1, response_format=response_format, **kwargs
            )
            for image in X
        ]


@dc.dataclass
class OpenAIAudioTranscription(OpenAI):
    """OpenAI audio transcription predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to guide the model's style. Should contain ``{context}``.
    """

    takes_context: bool = True
    prompt: str = ''

    @retry
    def _predict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Converts a file-like Audio recording to text."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            Audio.transcribe(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        )['text']

    @retry
    async def _apredict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Converts a file-like Audio recording to text."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            await Audio.atranscribe(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        )['text']

    @retry
    def _predict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Converts multiple file-like Audio recordings to text."
        resps = [
            Audio.transcribe(file=file, model=self.identifier, **kwargs)
            for file in files
        ]
        return [resp['text'] for resp in resps]

    @retry
    async def _apredict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Converts multiple file-like Audio recordings to text."
        resps = await asyncio.gather(
            *[
                Audio.atranscribe(file=file, model=self.identifier, **kwargs)
                for file in files
            ]
        )
        return [resp['text'] for resp in resps]

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        out = []
        batch_size = kwargs.pop('batch_size', 10)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        batch_size = kwargs.pop('batch_size', 10)
        list_of_lists = await asyncio.gather(
            *[
                self._apredict_a_batch(X[i : i + batch_size], **kwargs)
                for i in range(0, len(X), batch_size)
            ]
        )
        return list(itertools.chain(*list_of_lists))


@dc.dataclass
class OpenAIAudioTranslation(OpenAI):
    """OpenAI audio translation predictor.

    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to guide the model's style. Should contain ``{context}``.
    """

    takes_context: bool = True
    prompt: str = ''

    @retry
    def _predict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Translates a file-like Audio recording to English."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            Audio.translate(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        )['text']

    @retry
    async def _apredict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Translates a file-like Audio recording to English."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            await Audio.atranslate(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        )['text']

    @retry
    def _predict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Translates multiple file-like Audio recordings to English."
        resps = [
            Audio.translate(file=file, model=self.identifier, **kwargs)
            for file in files
        ]
        return [resp['text'] for resp in resps]

    @retry
    async def _apredict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Translates multiple file-like Audio recordings to English."
        resps = await asyncio.gather(
            *[
                Audio.atranslate(file=file, model=self.identifier, **kwargs)
                for file in files
            ]
        )
        return [resp['text'] for resp in resps]

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        out = []
        batch_size = kwargs.pop('batch_size', 10)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        batch_size = kwargs.pop('batch_size', 10)
        list_of_lists = await asyncio.gather(
            *[
                self._apredict_a_batch(X[i : i + batch_size], **kwargs)
                for i in range(0, len(X), batch_size)
            ]
        )
        return list(itertools.chain(*list_of_lists))



import copy
import dataclasses as dc
import io
import pickle
import types
import typing as t
from abc import ABC

import dill
import typing_extensions as te

Info = t.Optional[t.Dict[str, t.Any]]

_ATOMIC_TYPES = frozenset(
    {
        # Common JSON Serializable types
        type(None),
        bool,
        int,
        float,
        str,
        # Other common types
        complex,
        bytes,
        # Other types that are also unaffected by deepcopy
        type(...),
        type(NotImplemented),
        types.CodeType,
        types.BuiltinFunctionType,
        types.FunctionType,
        type,
        range,
        property,
    }
)


class ModuleClassDict(te.TypedDict):
    """
    A ``dict``with the module, class and JSONization of an object

    :param module: The module of the object
    """

    module: str
    cls: str
    dict: t.Dict[str, t.Any]


class Serializer(ABC):
    @staticmethod
    def encode(object: t.Any, info: Info = None) -> bytes:
        raise NotImplementedError

    @staticmethod
    def decode(b: bytes, info: Info = None) -> t.Any:
        raise NotImplementedError


class PickleSerializer(Serializer):
    @staticmethod
    def encode(object: t.Any, info: Info = None) -> bytes:
        return pickle.dumps(object)

    @staticmethod
    def decode(b: bytes, info: Info = None) -> t.Any:
        return pickle.loads(b)


class DillSerializer(Serializer):
    @staticmethod
    def encode(object: t.Any, info: Info = None) -> bytes:
        return dill.dumps(object)

    @staticmethod
    def decode(b: bytes, info: Info = None) -> t.Any:
        return dill.loads(b)


class TorchSerializer(Serializer):
    @staticmethod
    def encode(object: t.Any, info: Info = None) -> bytes:
        import torch

        from superduperdb.ext.torch.utils import device_of

        if not isinstance(object, dict):
            was_gpu = str(device_of(object)) == 'cuda'
            object.to('cpu')
            f = io.BytesIO()
            torch.save(object, f)
            if was_gpu:
                object.to('cuda')
        else:
            f = io.BytesIO()
            torch.save(object, f)
        return f.getvalue()

    @staticmethod
    def decode(b: bytes, info: Info = None) -> t.Any:
        import torch

        return torch.load(io.BytesIO(b))


class Serializers:
    serializers: t.Dict[str, t.Type] = {}

    def add(self, name: str, serializer: t.Type):
        self.serializers[name] = serializer

    def __getitem__(self, serializer):
        return self.serializers[serializer]


serializers = Serializers()
serializers.add('pickle', PickleSerializer)
serializers.add('dill', DillSerializer)
serializers.add('torch', TorchSerializer)


class Method:
    """
    A callable that calls a method on the object it is called with.

    :param method: The method to call.
    :param *args: The args to call the method with.
    :param **kwargs: The kwargs to call the method with.
    """

    def __init__(self, method: str, *args, **kwargs):
        self.method = method
        self.args = args
        self.kwargs = kwargs

    def __call__(self, X: t.Any) -> t.Any:
        return getattr(X, self.method)(*self.args, **self.kwargs)


def asdict(obj, *, copy_method=copy.copy) -> t.Dict[str, t.Any]:
    """
    Custom ``asdict`` function which exports a dataclass object into a dict,
    with a option to choose for nested non atomic objects copy strategy.
    """
    if not dc.is_dataclass(obj):
        raise TypeError("asdict() should be called on dataclass instances")
    return _asdict_inner(obj, dict, copy_method)


def _asdict_inner(obj, dict_factory, copy_method) -> t.Any:
    if type(obj) in _ATOMIC_TYPES:
        return obj
    elif dc.is_dataclass(obj):
        # fast path for the common case
        return {
            f.name: _asdict_inner(getattr(obj, f.name), dict, copy_method)
            for f in dc.fields(obj)
        }
    elif isinstance(obj, tuple) and hasattr(obj, '_fields'):
        # obj is a namedtuple.  Recurse into it, but the returned
        # object is another namedtuple of the same type.  This is
        # similar to how other list- or tuple-derived classes are
        # treated (see below), but we just need to create them
        # differently because a namedtuple's __init__ needs to be
        # called differently (see bpo-34363).

        # I'm not using namedtuple's _asdict()
        # method, because:
        # - it does not recurse in to the namedtuple fields and
        #   convert them to dicts (using dict_factory).
        # - I don't actually want to return a dict here.  The main
        #   use case here is json.dumps, and it handles converting
        #   namedtuples to lists.  Admittedly we're losing some
        #   information here when we produce a json list instead of a
        #   dict.  Note that if we returned dicts here instead of
        #   namedtuples, we could no longer call asdict() on a data
        #   structure where a namedtuple was used as a dict key.

        return type(obj)(*[_asdict_inner(v, dict_factory, copy_method) for v in obj])
    elif isinstance(obj, (list, tuple)):
        # Assume we can create an object of this type by passing in a
        # generator (which is not true for namedtuples, handled
        # above).
        return type(obj)(_asdict_inner(v, dict_factory, copy_method) for v in obj)
    elif isinstance(obj, dict):
        if hasattr(type(obj), 'default_factory'):
            # obj is a defaultdict, which has a different constructor from
            # dict as it requires the default_factory as its first arg.
            result = type(obj)(getattr(obj, 'default_factory'))
            for k, v in obj.items():
                result[_asdict_inner(k, dict_factory, copy_method)] = _asdict_inner(
                    v, dict_factory, copy_method
                )
            return result
        return type(obj)(
            (
                _asdict_inner(k, dict_factory, copy_method),
                _asdict_inner(v, dict_factory, copy_method),
            )
            for k, v in obj.items()
        )
    else:
        return copy_method(obj)



import warnings

from .configs import CFG

__all__ = ('logging',)


if CFG.logging.type == 'stdout':

    class logging:
        warn = staticmethod(warnings.warn)
        error = staticmethod(warnings.warn)
        debug = staticmethod(print)
        info = staticmethod(print)

    def dont_print(*a, **ka):
        pass

    level = CFG.logging.level

    if level == level.DEBUG:
        pass

    elif level == level.INFO:
        logging.debug = dont_print  # type: ignore[assignment]

    else:
        logging.debug = dont_print  # type: ignore[assignment]
        logging.info = dont_print  # type: ignore[assignment]

else:
    import logging  # type: ignore[assignment]

    level = getattr(logging, CFG.logging.level.name)
    logging.basicConfig(level=level, **CFG.logging.kwargs)  # type: ignore[attr-defined]
    logging.getLogger('distributed').propagate = True  # type: ignore[attr-defined]
    logging.getLogger('vcr').setLevel(logging.WARNING)  # type: ignore[attr-defined]



import os
import typing as t


def str_shape(shape: t.Sequence[int]) -> str:
    if not shape:
        raise ValueError('Shape was empty')
    return 'x'.join(str(x) for x in shape)


def get_key(key_name: str) -> str:
    try:
        return os.environ[key_name]
    except KeyError:
        raise KeyError(f'Environment variable {key_name} is not set') from None


def format_prompt(X: str, prompt: str, context: t.Optional[t.List[str]] = None) -> str:
    if '{context}' in prompt:
        if context:
            return prompt.format(context='\n'.join(context)) + X
        else:
            raise ValueError(f'A context is required for prompt {prompt}')
    else:
        return prompt + X



"""
Change Data Capture (CDC) is a mechanism used in database systems to track
and capture changes made to a table or collection in real-time.
It allows applications to stay up-to-date with the latest changes in the database
and perform various tasks, such as data synchronization, auditing,
or data integration. The ChangeDataCapture class is designed to
handle CDC tasksfor a specified table/collection in a database.

Change streams allow applications to access real-time data changes
without the complexity and risk of tailing the oplog.
Applications can use change streams to subscribe to all data
changes on a single collection,a database, or an entire deployment,
and immediately react to them.
Because change streams use the aggregation framework, applications
can also filter for specific changes or transform the notifications at will.

ref: https://www.mongodb.com/docs/manual/changeStreams/

Use this module like this::
    db = any_arbitary_database.connect(...)
    db = superduper(db)
    listener = DatabaseListener(db=db, on=Collection('test_collection'))
    listener.listen()
"""

import typing as t

from superduperdb.db.base import backends
from superduperdb.db.base.db import DB
from superduperdb.db.mongodb.cdc.base import BaseDatabaseListener
from superduperdb.db.mongodb.cdc.db_listener import MongoDatabaseListener
from superduperdb.db.mongodb.query import Collection
from superduperdb.misc.runnable.runnable import Event

DBListenerType = t.TypeVar('DBListenerType')


class DatabaseListenerFactory(t.Generic[DBListenerType]):
    """A Factory class to create instance of DatabaseListener corresponding to the
    `db_type`.
    """

    SUPPORTED_LISTENERS: t.List[str] = ['mongodb']

    def __init__(self, db_type: str = 'mongodb'):
        if db_type not in self.SUPPORTED_LISTENERS:
            raise ValueError(f'{db_type} is not supported yet for CDC.')
        self.listener = db_type

    def create(self, *args, **kwargs) -> DBListenerType:
        stop_event = Event()
        kwargs['stop_event'] = stop_event
        listener = MongoDatabaseListener(*args, **kwargs)
        return t.cast(DBListenerType, listener)


def DatabaseListener(
    db: DB,
    on: Collection,
    identifier: str = '',
    *args,
    **kwargs,
) -> BaseDatabaseListener:
    """
    Create an instance of ``BaseDatabaseListener``.
    Not to be confused with ``superduperdb.container.listener.Listener``.

    :param db: A superduperdb instance.
    :param on: Which collection/table listener service this be invoked on?
    :param identifier: A identity given to the listener service.
    """
    it = backends.data_backends.items()
    if types := [k for k, v in it if isinstance(db.databackend, v)]:
        db_type = types[0]
    else:
        raise ValueError('No backends found')

    if db_type != 'mongodb':
        raise NotImplementedError(f'Database {db_type} not supported yet!')

    factory_factory = DatabaseListenerFactory[MongoDatabaseListener]
    db_factory = factory_factory(db_type=db_type)
    return db_factory.create(db=db, on=on, identifier=identifier, *args, **kwargs)



"""
TODO:
- select_ids_of_missing_outputs
"""

import dataclasses as dc
import enum
import re
import typing as t

import ibis

from superduperdb.base.logger import logging
from superduperdb.container.component import Component
from superduperdb.container.encoder import Encoder
from superduperdb.container.schema import Schema
from superduperdb.db.base.cursor import SuperDuperCursor
from superduperdb.db.base.db import DB
from superduperdb.db.base.query import (
    CompoundSelect,
    Insert,
    Like,
    QueryComponent,
    QueryLinker,
    Select,
    TableOrCollection,
    _ReprMixin,
)

PRIMARY_ID: str = 'id'

IbisTableType = t.TypeVar('IbisTableType')
ParentType = t.TypeVar('ParentType')


@dc.dataclass(repr=False)
class IbisCompoundSelect(CompoundSelect):
    """
    A query incorporating vector-search and a standard ``ibis`` query
    """

    __doc__ = __doc__ + CompoundSelect.__doc__  # type: ignore[operator]

    def __eq__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__eq__(__value)

    def __lt__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__lt__(__value)

    def __gt__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__gt__(__value)

    def __and__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__and__(__value)

    def __or__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__or__(__value)

    def __not__(self) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__not__()

    def __getitem__(self, item) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__getitem__(item)

    def _get_query_linker(cls, table_or_collection, members) -> 'IbisQueryLinker':
        return IbisQueryLinker(table_or_collection=table_or_collection, members=members)

    def _get_query_component(
        self,
        name: str,
        type: str,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return IbisQueryComponent(name, type=type, args=args, kwargs=kwargs)

    def outputs(self, model: str):
        """
        This method returns a query which joins a query with the outputs
        for a table.

        :param model: The model identifier for which to get the outputs

        >>> q = t.filter(t.age > 25).outputs('model_name', db)

        The above query will return the outputs of the `model_name` model
        with t.filter() ids.
        """
        assert self.query_linker is not None
        return IbisCompoundSelect(
            table_or_collection=self.table_or_collection,
            pre_like=self.pre_like,
            query_linker=self.query_linker.outputs(model),
            post_like=self.post_like,
        )

    def compile(self, db: DB, tables: t.Optional[t.Dict] = None):
        """
        Convert the current query to an ``ibis`` native query.

        :param db: The superduperdb connection
        :param tables: A dictionary of ``ibis`` tables to use for the query
        """
        assert self.pre_like is None, "pre_like must be None"
        assert self.post_like is None, "post_like must be None"
        assert self.query_linker is not None, "query_linker must be set"

        table_id = self.table_or_collection.identifier
        if tables is None:
            tables = {}
        if table_id not in tables:
            tables[table_id] = db.databackend.conn.table(table_id)
        return self.query_linker.compile(db, tables=tables)

    def get_all_tables(self):
        tables = [self.table_or_collection.identifier]
        tmp = self.query_linker.get_all_tables()
        tables.extend(tmp)
        tables = list(set(tables))
        return tables

    def _get_all_fields(self, db):
        tables = self.get_all_tables()
        component_tables = []
        for tab in tables:
            component_tables.append(db.load('table', tab))
        fields = {}
        for tab in component_tables:
            fields.update(tab.schema.fields)
        return fields

    @property
    def select_table(self):
        return self.table_or_collection

    def execute(self, db):
        output, scores = super().execute(db)
        fields = self._get_all_fields(db)
        for column in output.columns:
            try:
                type = fields[column]
            except KeyError:
                logging.warn(f'Disambiguation not yet supported of {column}: TODO!')
                continue
            if isinstance(type, Encoder):
                output[column] = output[column].map(type.decode)
        if scores is not None:
            output['scores'] = list(
                map(scores, output[self.table_or_collection.primary_id].tolist())
            )
        output = output.to_dict(orient='records')
        return SuperDuperCursor(output, id_field=self.table_or_collection.primary_id)


class _LogicalExprMixin:
    def _logical_expr(self, members, collection, k, other: t.Optional[t.Any] = None):
        if other:
            args = [other]
        else:
            args = []
        members.append(
            IbisQueryComponent(k, args=args, kwargs={}, type=QueryType.QUERY)
        )
        return IbisQueryLinker(collection, members=members)

    def eq(self, other, members, collection):
        k = '__eq__'
        return self._logical_expr(members, collection, k, other=other)

    def or_(self, other, members, collection):
        k = '__or__'
        return self._logical_expr(members, collection, k, other=other)

    def not_(self, members, collection):
        k = '__not__'
        return self._logical_expr(members, collection, k)

    def and_(self, other, members, collection):
        k = '__and__'
        return self._logical_expr(members, collection, k, other=other)

    def gt(self, other, members, collection):
        k = '__gt__'
        return self._logical_expr(members, collection, k, other=other)

    def lt(self, other, members, collection):
        k = '__lt__'
        return self._logical_expr(other, members, collection, k, other=other)

    def getitem(self, other, members, collection):
        k = '__getitem__'
        return self._logical_expr(members, collection, k, other=other)


@dc.dataclass(repr=False)
class IbisQueryLinker(QueryLinker, _LogicalExprMixin):
    def repr_(self) -> str:
        out = super().repr_()
        out = re.sub('\. ', ' ', out)
        return out

    def __eq__(self, other):
        return self.eq(other, members=self.members, collection=self.table_or_collection)

    def __lt__(self, other):
        return self.lt(other, members=self.members, collection=self.table_or_collection)

    def __gt__(self, other):
        return self.gt(other, members=self.members, collection=self.table_or_collection)

    def __or__(self, other):
        return self.or_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __and__(self, other):
        return self.and_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __not__(self, other):
        return self.not_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __getitem__(self, other):
        return self.getitem(
            other, members=self.members, collection=self.table_or_collection
        )

    def _get_query_component(self, k):
        return IbisQueryComponent(name=k, type=QueryType.ATTR)

    @property
    def select_ids(self):
        return self.select(self.table_or_collection.primary_id)

    def select_single_id(self, id):
        return self.filter(
            self.table_or_collection.__getattr__(self.table_or_collection.primary_id)
            == id
        )

    def select_using_ids(self, ids):
        return self.filter(
            self.table_or_collection.__getattr__(
                self.table_or_collection.primary_id
            ).isin(ids)
        )

    def select_ids_of_missing_outputs(self, key: str, model: str):
        raise NotImplementedError

    def get_all_tables(self):
        out = []
        for member in self.members:
            out.extend(member.get_all_tables())
        return list(set(out))

    def outputs(self, model: str):
        symbol_table = IbisQueryTable(
            identifier=f'_outputs/{model}',
            primary_id='id',
        )
        attr = getattr(  # type: ignore[call-overload]
            self, self.table_or_collection.primary_id
        )
        other_query = self.join(
            symbol_table,
            [
                symbol_table.input_id == attr,
                symbol_table.query_id == self.table_or_collection.identifier,
            ],
        )
        return other_query

    def compile(self, db: DB, tables: t.Optional[t.Dict] = None):
        table_id = self.table_or_collection.identifier
        if tables is None:
            tables = {}
        if table_id not in tables:
            tables = {table_id: db.databackend.conn.table(table_id)}
        result = tables[table_id]
        for member in self.members:
            result, tables = member.compile(parent=result, db=db, tables=tables)
        return result, tables

    def execute(self, db):
        native_query, _ = self.compile(db)
        result = native_query.execute()
        return result


class QueryType(str, enum.Enum):
    QUERY = 'query'
    ATTR = 'attr'


@dc.dataclass(repr=False)
class IbisTable(Component):
    """
    This is a representation of an SQL table in ibis,
    saving the important meta-data associated with the table
    in the ``superduperdb`` meta-data store.

    :param identifier: The name of the table
    :param schema: The schema of the table
    :param primary_id: The primary id of the table
    :param version: The version of the table
    """

    identifier: str
    schema: Schema
    primary_id: str = 'id'
    version: t.Optional[int] = None
    type_id: t.ClassVar[str] = 'table'

    def on_create(self, db: DB):
        assert self.schema is not None, "Schema must be set"
        for e in self.schema.encoders:
            db.add(e)
        if db.databackend.in_memory:
            logging.info(f'Using in-memory tables "{self.identifier}" so doing nothing')
            return

        try:
            db.databackend.create_ibis_table(  # type: ignore[attr-defined]
                self.identifier, schema=ibis.schema(self.schema.raw)
            )
        except Exception as e:
            if 'already exists' in str(e):
                pass
            else:
                raise e

    def insert(self, documents, **kwargs):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).insert(documents, **kwargs)

    def __getattr__(self, item):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).__getattr__(item)

    def __getitem__(self, item):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).__getitem__(item)


@dc.dataclass(repr=False)
class IbisQueryTable(_ReprMixin, TableOrCollection, Select):
    """
    This is a symbolic representation of a table
    for building ``IbisCompoundSelect`` queries.

    :param identifier: The name of the table
    :param primary_id: The primary id of the table
    :param version: The version of the table
    """

    primary_id: str = 'id'

    def compile(self, db: DB, tables: t.Optional[t.Dict] = None):
        if tables is None:
            tables = {}
        if self.identifier not in tables:
            tables[self.identifier] = db.databackend.conn.table(self.identifier)
        return tables[self.identifier], tables

    def repr_(self):
        return self.identifier

    def add_fold(self, fold: str) -> Select:
        return self.filter(self.fold == fold)

    @property
    def id_field(self):
        return self.primary_id

    @property
    def select_table(self) -> Select:
        return self

    @property
    def select_ids(self) -> Select:
        return self.select(self.primary_id)

    def select_using_ids(self, ids: t.Sequence[t.Any]) -> Select:
        return self.filter(self[self.primary_id].isin(ids))

    def select_ids_of_missing_outputs(self, key: str, model: str) -> Select:
        raise NotImplementedError

    def select_single_id(self, id):
        return self.filter(getattr(self, self.primary_id) == id)

    def __getitem__(self, item):
        return IbisCompoundSelect(
            table_or_collection=self,
            query_linker=self._get_query_linker(
                members=[IbisQueryComponent('__getitem__', type=QueryType.ATTR)]
            ),
        )(item)

    def _insert(self, documents, **kwargs):
        return IbisInsert(documents=documents, kwargs=kwargs, table_or_collection=self)

    def _get_query(
        self,
        pre_like: t.Optional[Like] = None,
        query_linker: t.Optional[QueryLinker] = None,
        post_like: t.Optional[Like] = None,
    ) -> IbisCompoundSelect:
        return IbisCompoundSelect(
            pre_like=pre_like,
            query_linker=query_linker,
            post_like=post_like,
            table_or_collection=self,
        )

    def _get_query_component(
        self,
        k,
    ):
        return IbisQueryComponent(name=k, type=QueryType.ATTR)

    def _get_query_linker(self, members) -> IbisQueryLinker:
        return IbisQueryLinker(table_or_collection=self, members=members)

    def model_update(  # type: ignore[override]
        self,
        db,
        ids: t.Sequence[t.Any],
        key: str,
        model: str,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
    ):
        if flatten:
            raise NotImplementedError('Flatten not yet supported for ibis')

        if key.startswith('_outputs'):
            key = key.split('.')[1]

        if not outputs:
            return

        # TODO generalize this to multiple queries per table/ model
        table_records = []
        for ix in range(len(outputs)):
            d = {
                'id': str(ids[ix]),
                'input_id': str(ids[ix]),
                'query_id': self.identifier,
                'output': outputs[ix],
                'key': key,
            }
            table_records.append(d)

        for r in table_records:
            if isinstance(r['output'], dict) and '_content' in r['output']:
                r['output'] = r['output']['_content']['bytes']

        db.databackend.insert(f'_outputs/{model}', table_records)

    def insert(
        self,
        *args,
        **kwargs,
    ):
        return self._insert(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return super()._delete(*args, **kwargs)

    def execute(self, db):
        return db.databackend.conn.table(self.identifier).execute()


def _compile_item(item, db, tables):
    if isinstance(item, (IbisQueryLinker, IbisCompoundSelect, IbisQueryTable)):
        return item.compile(db, tables=tables)
    if isinstance(item, list) or isinstance(item, tuple):
        compiled = []
        for x in item:
            c, tables = _compile_item(x, db, tables=tables)
            compiled.append(c)
        return compiled, tables
    elif isinstance(item, dict):
        c = {}
        for k, v in item.items():
            c[k], tables = _compile_item(v, db, tables=tables)
        return c, tables
    return item, tables


def _get_all_tables(item):
    if isinstance(item, (IbisQueryLinker, IbisCompoundSelect)):
        return item.get_all_tables()
    elif isinstance(item, IbisQueryTable):
        return [item.identifier]
    elif isinstance(item, list) or isinstance(item, tuple):
        return sum([_get_all_tables(x) for x in item], [])
    elif isinstance(item, dict):
        return sum([_get_all_tables(x) for x in item.values()], [])
    else:
        return []


@dc.dataclass
class IbisQueryComponent(QueryComponent):
    """
    This class represents a component of an ``ibis`` query.
    For example ``filter`` in ``t.filter(t.age > 25)``.
    """

    __doc__ = __doc__ + QueryComponent.__doc__  # type: ignore[operator]

    def repr_(self) -> str:
        out = super().repr_()
        match = re.match('.*__([a-z]{2,2})__\(([a-z0-9_\.\']+)\)', out)
        lookup = {'gt': '>', 'lt': '<', 'eq': '=='}
        if match is not None and match.groups()[0] in lookup:
            out = f' {lookup[match.groups()[0]]} {match.groups()[1]}'
        return out

    def compile(self, parent: t.Any, db: DB, tables: t.Optional[t.Dict] = None):
        if self.type == QueryType.ATTR:
            return getattr(parent, self.name), tables
        args, tables = _compile_item(self.args, db, tables=tables)
        kwargs, tables = _compile_item(self.kwargs, db, tables=tables)
        return getattr(parent, self.name)(*args, **kwargs), tables

    def get_all_tables(self):
        out = []
        out.extend(_get_all_tables(self.args))
        out.extend(_get_all_tables(self.kwargs))
        return list(set(out))


@dc.dataclass
class IbisInsert(Insert):
    def _encode_documents(self, table: IbisTable) -> t.List[t.Dict]:
        return [r.encode(table.schema) for r in self.documents]

    def execute(self, db):
        table = db.load(
            'table',
            self.table_or_collection.identifier,
        )
        encoded_documents = self._encode_documents(table=table)
        ids = [r[table.primary_id] for r in encoded_documents]
        db.databackend.insert(
            self.table_or_collection.identifier, raw_documents=encoded_documents
        )
        return ids

    @property
    def select_table(self):
        return self.table_or_collection
